# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

from Frontend.C import c_parser, c_ast


from Backends.Cuda.Visitors.CM_Visitors import *

from Tools.Tree import InsertTool, NodeNotFound, ReplaceTool, RemoveTool

from Tools import SourceStorage

from Backends.Cuda.Mutators.Common import AbstractCudaMutator

class CM_OmpFor(AbstractCudaMutator):

    def filter(self, ast):
        """ Filter definition
            Returns the first node matching with the filter
        """
        # Build a visitor , matching the OmpFor node of the AST
        f = OmpForFilter()
        node = f.apply(ast)
        self._func_def = f.get_func_def()
        self._parallel = f.get_parallel()
        return node

    def apply_all(self, parent_parallel_node, ast):
        """ Apply mutation to all matches """
        start_node = None
        self.ast = ast
        f = OmpForFilter()
        num = 0;
        try:
            for elem in f.iterate(ast):
                # Save previous state
                old_name = self.kernel_name
                old_clauses = self._clauses
                self.kernel_name = self.kernel_name + str(num)
                # Current scope variables
                self._func_def = f.get_func_def()
                self._parallel = f.get_parallel()
                # If current node is not child of first parallel node, stop
                if self._parallel != parent_parallel_node:
                    raise StopIteration
                self.source_storage.addFile(self.kernel_name, SourceStorage.CUDA_SOURCE)
                self.source_storage.append(self.kernel_name, SourceStorage.CUDA_SOURCE, '#include "llcomp_cuda.h"\n')
                start_node = self.mutatorFunction(ast, elem)
                # Restore previous state
                self.kernel_name = old_name
                self._clauses = old_clauses
                num += 1;
        except NodeNotFound as nf:
            print str(nf)
        except StopIteration:
            return self._parallel
        return start_node

    def mutatorFunction(self, ast, ompFor_node):
        """ CUDA mutator, writes memory transfer operations for a parallel region
        """
        maxThreadNumber_node = self.getThreadNum(ompFor_node.stmt.cond)

        ##################### Statement for cuda
        cuda_stmts = c_ast.Compound(stmts = [], decls = []);
        ##################### Cuda parameters on host
        clause_dict = self._get_dict_from_clauses(ompFor_node.clauses,  ast)
  
        reduction_params = clause_dict['REDUCTION']
        nowait = clause_dict.has_key('NOWAIT')
        # Check if nested for
        collapse = clause_dict.has_key('COLLAPSE')
        # Private declarations come from the parent parallel construct
        private_params = clause_dict['PRIVATE']
        shared_params = clause_dict['SHARED']


        ##################### Declarations
#        kernel_init_subtree = self.buildDeclarations(numThreads = maxThreadNumber_node, reduction_node_list = reduction_params, shared_node_list = shared_params, device_node_list = [], ast = ast)
#        InsertTool(subtree = kernel_init_subtree, position = "begin").apply(cuda_stmts, 'decls')

        ##################### Kernel Call / Initialization

        # This pragma represents a parallel region, so we need to transfer shared data to device
        [externKernelDef_subtree, kernelCall_subtree] = self.buildKernelCall(
                                numThreads = maxThreadNumber_node, 
                                reduction_node_list = reduction_params, 
                                shared_node_list = shared_params, 
                                device_node_list = shared_params, ast = ast)
        # subtree needs to be a node list
        InsertTool(subtree = [externKernelDef_subtree], position = "begin").apply(ast, 'ext')
        InsertTool(subtree = [kernelCall_subtree], position = "end").apply(cuda_stmts, 'stmts')


        ##################### Cuda Kernel 
        # Kernel
        kernel_subtree = self.buildKernel(clause_dict = clause_dict, loop = ompFor_node.stmt, ast = ast)
        # Function declaration
        # TODO: Issue #7 bitbucket
        # - Build a node without body
#        tmp = c_ast.CUDAKernel(function = copy.deepcopy(kernel_subtree.ext[0].function), type = 'global', name = kernel_subtree.ext[0].name)
#        tmp.function.body = c_ast.Compound(stmts = None, decls = None); # If both of stmts and decls are none, it won't be printed
#        kernel_decl = c_ast.Compound(stmts = [tmp], decls = None)
#        # Find container function
#        InsertTool(subtree = kernel_decl, position = "begin", node = self._func_def).apply(ast, 'ext')
        # Function definition
        # InsertTool(subtree = kernel_subtree, position = "begin", node = self._func_def ).apply(ast, 'ext')

        ##################### Loop substitution 
        # Kernel Launch
#        kernelLaunch_subtree = self.buildKernelLaunch(reduction_vars = reduction_params, shared_vars = shared_params, ast = ompFor_node)
#        InsertTool(subtree = kernelLaunch_subtree, position = "end").apply(cuda_stmts, 'stmts')
        # Retrieve data
#        retrieve_subtree = self.buildRetrieve(reduction_vars = reduction_params, modified_shared_vars = shared_params)
#        InsertTool(subtree = retrieve_subtree, position = "end").apply(cuda_stmts, 'stmts')
#        # Host reduction
#        reduction_subtree = self.buildHostReduction(reduction_vars = reduction_params, ast = ast)
#        InsertTool(subtree = reduction_subtree, position = "end").apply(cuda_stmts, 'stmts')

        # Replace the entire pragma by a CompoundStatement with all the new statements
#        DotDebugTool(highlight = [ompFor_node]).apply(self._parallel)
        if isinstance(ompFor_node.parent.parent, c_ast.Compound) :
            ReplaceTool(new_node = cuda_stmts, old_node = ompFor_node.parent).apply(ompFor_node.parent.parent, 'stmts')
        else:
            # Maybe we have an error here
            ompFor_node.parent.parent.stmt = cuda_stmts




