# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------


from Backends.C99.Visitors.OmpFilters import OmpParallelFilter, OmpThreadPrivateFilter

from Tools.Tree import InsertTool, NodeNotFound, ReplaceTool

from Tools.Declarations import decl_of_id

from Backends.Cuda.Mutators.Common import AbstractCudaMutator

from Backends.Common.TemplateEngine.TemplateParser import get_template_array

from Tools import SourceStorage

class FM_LlcRegion(AbstractCudaMutator):
    def filter(self, ast):
        """ Filter definition
            Returns the first node matching with the filter
        
        """
        # Build a visitor , matching the OmpFor node of the AST
        f = OmpParallelFilter(device = self.device)
        node = f.apply(ast)
        self._func_def = f.get_func_def()
        self._parallel = node
        return node

    def apply_all(self, ast):
        """ Apply mutation to all matches """
        self.ast = ast
        f = OmpParallelFilter(device = self.device)
        num = 0;
        self.kernel_name = self.kernel_prefix
        old_prefix = self.kernel_prefix
        old_parallel = self._parallel
        try:
            for elem in f.iterate(ast):
                # Save previous state
                self.kernel_name = self.kernel_name + str(num)
                self.kernel_prefix = self.kernel_prefix + str(num)
                self._clauses = {}
                # Current scope variables
                self._func_def = f.get_func_def()
                self._target_device_node = f._target_device_node
                self._clauses = self._get_dict_from_clauses(f._target_device_node.clauses, ast)
                self._parallel = elem
                self.source_storage.addFile(self.kernel_name, SourceStorage.CUDA_SOURCE)
                self.source_storage.append(self.kernel_name, SourceStorage.CUDA_SOURCE, '#include "llcomp_cuda.h" \n')
                self.mutatorFunction(ast, elem)
                # Restore previous state
                # TODO : Maybe a parent relink is needed?
                self._parallel = old_parallel
                self.kernel_prefix = old_prefix
                self.kernel_name = old_prefix
                num += 1;
        except NodeNotFound as nf:
            print str(nf)
        except StopIteration:
            return self._parallel
        return ast



    def buildParallelRegionCreation(self, shared_node_list, ast):
        """ Builds the declaration section of a Parallel Region
                This code handles memory transfers between host and cuda
              
             :param shared_node_list: List of shared variable declarations
             :param ast: Original ast (for type search)

             :return: Parallel Declarations subtree
        """ 
        shared_vars = get_template_array(shared_node_list, ast) 
        template_code = """
            int main() {
                  <%namespace name="declarations" file="device_declaration.tmpl"/>
                  ${declarations.write_declarations(shared_vars, '', '_cu')}     
                  ${declarations.write_allocate_to_device(shared_vars, '', '_cu')}     

                 <%namespace name="transfers" file="device_transfers.tmpl"/>
                 ${transfers.write_transfer_to_device(shared_vars, '', '_cu')}     
            }

            """
        print "New kernel build from parallel region with name : " + self.kernel_name
        parallel_init = self.parse_snippet(template_code, {'shared_vars' : shared_vars}, name = 'Initialization of Parallel Region ' + self.kernel_name, show = False).ext[-1].body
        return parallel_init



    def buildParallelRegionDestroy(self, reduction_vars, modified_shared_vars, ast = None, shared_vars = None):
        reduction_vars = get_template_array(reduction_vars, ast)
        shared_vars    = get_template_array(modified_shared_vars, ast)

        # Template source
        template_code = """
        int fake() {
        <%namespace name="snippets" file="snippets.tmpl"/>
        ${snippets.write_retrieve_all_data_from_device(reduction_vars, shared_vars)}     
        }
        """ 
        return self.parse_snippet(template_code, {'reduction_vars' : reduction_vars, 'shared_vars' : shared_vars}, name = 'Retrieve', show = False).ext[0].body

    def mutatorFunction(self, ast, ompParallel_node):
        """ CUDA mutator, writes memory transfer operations for a parallel region
        """
        from Backends.Cuda.Mutators.CM_llcNestedFor import CM_llcNestedFor
        from Backends.Cuda.Mutators.CM_OmpFor import CM_OmpFor

        threadprivate = []
        for elem in OmpThreadPrivateFilter().dfs_iter(ast):
            threadprivate.extend([decl_of_id(it, ast) for it in elem.identifiers.params])
        # print " Threadprivate : " + str(threadprivate)

        clause_dict = self._get_dict_from_clauses(ompParallel_node.clauses, ast)

        shared_params = clause_dict['SHARED']
        modified_shared_vars = clause_dict['COPY_OUT']
        copyin_shared_vars = clause_dict['COPY_IN']

        # If no copy clauses specified, default to shared params
        if len(copyin_shared_vars) == 0 and len(modified_shared_vars) == 0:
            modified_shared_vars = shared_params
            copyin_shared_vars = shared_params

        private_params = clause_dict['PRIVATE'] 
        nowait = clause_dict.has_key('NOWAIT')
        # If the parallel statement have declarations, they are private to the thread, so, we need to put them as params
        if ompParallel_node.stmt:
            if ompParallel_node.stmt.decls:
                private_params += ompParallel_node.stmt.decls
        private_params.extend(threadprivate)

        # Loops inside parallel region (wired for now)
        CM_llcNestedFor(clause_dict, kernel_name = self.kernel_prefix + "_nestedLoopKernel",  source_storage = self.source_storage).apply_all(ompParallel_node, ast)
        CM_OmpFor(clause_dict, kernel_name = self.kernel_prefix + "_loopKernel", source_storage = self.source_storage).apply_all(ompParallel_node, ast)

        ##################### Statement for cuda
        from Frontend.C import c_ast
        cuda_stmts = c_ast.Compound(stmts = [], decls = []);
        ##################### Cuda parameters on host
        # Declarations
        declarations_subtree = self.buildParallelRegionCreation(shared_node_list = copyin_shared_vars, ast = ast)
        InsertTool(subtree = declarations_subtree, position = "begin").apply(cuda_stmts, 'decls')
        # Initialization
        InsertTool(subtree = self._parallel.stmt, position = "begin").apply(cuda_stmts, 'stmts')
        # Retrieve data
        retrieve_subtree = self.buildParallelRegionDestroy(reduction_vars = [], modified_shared_vars = modified_shared_vars, ast = ast, shared_vars = copyin_shared_vars)
        InsertTool(subtree = retrieve_subtree, position = "end").apply(cuda_stmts, 'stmts')

        ##################### Loop substitution 
        # Replace the entire pragma by a CompoundStatement with all the new statements
        # Note: The parent of Parallel is always a Pragma node
        ReplaceTool(new_node = cuda_stmts, old_node = self._parallel.parent).apply(self._parallel.parent.parent, 'stmts')


