# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

from Backends.Cuda.Mutators.Common import AbstractCudaMutator
from Backends.C99.Visitors.OmpFilters import OmpParallelForFilter

from Frontend.Shortcuts import getCurrentLanguageAst
c_ast = getCurrentLanguageAst()

from Tools import SourceStorage
from Tools.Tree import InsertTool, NodeNotFound, ReplaceTool
from Backends.Cuda.Files.FileTypes import Cuda_FileType
from Backends.Cuda.Kernelize import cuda_outline_loop_stmts


class CM_OmpParallelFor(AbstractCudaMutator):
    """ This  mutator locates a omp parallel for reduction, and then
        translate the original source to an equivalent cuda implementation 

    """
    def __init__(self, clauses = None, kernel_name = 'loopKernel', kernel_prefix = '', *args, **kwargs):
        """ Constructor """
        self._ompFor = None
        self._clauses = {}
        self._backend_name = "Cuda"
        self.kernel_name = kernel_name
        super(CM_OmpParallelFor, self).__init__(clauses, kernel_name, kernel_prefix, *args, **kwargs)


    def filter(self, ast):
        """ Filter definition
            
        @return first node matching with the filter
        """
        # Build a visitor , matching the OmpFor node of the AST
        f = OmpParallelForFilter(device = 'cuda')
        node = f.apply(ast)
        self._func_def = f.get_func_def()
        self._parallel = f.get_parallel()
        return node

    def apply_all(self, ast):
        """ Apply mutation to all matches 
            
            @return last node changed
        """
        start_node = None
        self.ast = ast
        f = OmpParallelForFilter(device = 'cuda')
        num = 0;
        self.kernel_name = self.kernel_prefix + self.kernel_name
        old_prefix = self.kernel_prefix
        old_parallel = self._parallel

        try:
            for elem in f.iterate(ast):
                # Save previous state
                old_name = self.kernel_name
                old_clauses = self._clauses
                self.kernel_name = self.kernel_name + str(num)
                self.kernel_prefix = self.kernel_prefix + str(num)
                # Current scope variables
                self._func_def = f.get_func_def()
                self._parallel = f.get_parallel()
                self._ompFor = f.get_ompFor()
                # Apply mutation
                start_node = self.mutatorFunction(ast, self._ompFor)
                # Restore previous state
                self.kernel_name = old_name
                self._clauses = old_clauses
                num += 1;
        except NodeNotFound as nf:
            print " This is the one "
            print str(nf)
        except StopIteration:
            return self._parallel

        return start_node




    def mutatorFunction(self, ast, ompFor_node):
        """ Main mutator for OpenMP Parallel For construct

            Writes the optimized code of an OpenMP Parallel For construct, building a kernel
                overwriting the for loop.
        """
        from Tools.Tree import getFileNode
        self.main_ast = getFileNode(ompFor_node)
        self._backend_name = "Cuda"
        kernel_source = cuda_outline_loop_stmts(ompFor_node.stmt.stmt, kernel_name = self.kernel_name)
        
        # Add a new file per each kernel
        self.source_storage.addFile(self.kernel_name, Cuda_FileType())
        self.source_storage.append(self.kernel_name, Cuda_FileType(), '#include "llcomp_cuda.h" \n')
        self.source_storage.append(self.kernel_name, Cuda_FileType(), kernel_source)
 
