# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

# from Frontend.C import c_ast

from Frontend.Shortcuts import getCurrentLanguageAst

c_ast = getCurrentLanguageAst

from Backends.Common.Visitors.GenericVisitors import *


from Tools.Tree import InsertTool, NodeNotFound, ReplaceTool, RemoveTool
from Tools.Declarations import type_of_id, decl_of_id
from Tools import SourceStorage

from Tools.Debug import DotDebugTool
# from Frontend.Parse import parse_source
from Backends.Common.Mutators.AstSupport import IDNameMutator
from Backends.Common.Mutators.AbstractMutator import IgnoreMutationException, AbstractMutator, AbortMutationException

from Backends.Common.TemplateEngine.TemplateParser import BackendTemplateParser, TemplateParser, get_template_array, get_typedefs_to_template

import config

from Tools import SourceStorage

from MiddleEnd.Functions.Mutators import Outliner




class CudaMutatorError(Exception):
    def __init__(self, description):
        self.description = description

    def __str__(self):
        return "CudaMutatorError :: " + self.description


class AbstractCudaMutator(Outliner.OutlineMutator):
    """ Common methods to work with CUDA

    """
    def __init__(self, clauses = None, kernel_name = 'loopKernel', kernel_prefix = '', source_storage = None):
        if clauses == None:
            clauses = {}
        self._kernel_name = kernel_name
        self.kernel_prefix = "CM_" + kernel_prefix
        self._func_def = None
        self._parallel = None
        self._clauses = clauses
        self.device = "cuda"
        self._backend_name = 'Cuda'
        self.source_storage = source_storage

        super(AbstractCudaMutator, self, ).__init__()

    # TODO: Clean this function, it has an strange behavior
    # TODO: Change function name
    #===========================================================================
    # def get_names(self, elem, ast):
    #    """ Return a list of names for a type """
    #    type = type_of_id(elem, ast)
    #    while not hasattr(type, 'names') and not hasattr(type, 'name'):
    #        # TODO: Launch exception if not type attribute
    #        type = type.type
    #    if hasattr(type, 'names'):
    #        return type.names
    #    else:
    #        return [type.name]
    #===========================================================================


	 # TODO: Change method name to getLoopSize
    def getThreadNum(self, node):
        """ Gets the maximum number of threads needed """
        if node.op == '<' or node.op == '<=':
             return node.right
        else:
             return node.left

 
    def buildKernelCall(self, numThreads, clause_dict, function_body):
        """ Builds the declaration section 

            :param numThreads: number of threads to launch
            :param reduction_node_list: Declarations of the reduction clause
            :param shared_node_list: Declarations of the shared clause
            :param device_node_list: Declarations of vars going to device
            :return: Declarations subtree
        """ 
        # Position in the template for dimA declaration, just in case we change it
        reduction_vars = get_template_array(clause_dict['REDUCTION'], self.main_ast,)
        shared_vars = get_template_array(clause_dict['SHARED'], self.main_ast,) 
        device_vars = get_template_array(clause_dict['PRIVATE'], self.main_ast,)

        clause_vars = []
        clause_vars.extend(shared_vars)
        clause_vars.extend(device_vars)
        typedef_list = get_typedefs_to_template(clause_vars, self.main_ast)

        # TODO: Use TemplateLookup instead of this
        template_code = open(config.WORKDIR + '/Backends/Cuda/Templates/kernel_call.tmpl', 'r').read()
        # print "*****>>> " + str(self._backend_name)
        template_code = BackendTemplateParser(self._backend_name, template_code).render(**{'dimA': str(numThreads) , 
                                                                                   'kernelName' : self._kernel_name, 'device_vars': device_vars, 'reduction_vars' : reduction_vars, 'shared_vars' : shared_vars, 'typedefs' : typedef_list})
        self.source_storage.append(self._kernel_name, SourceStorage.CUDA_SOURCE, template_code)
    
        [kernel_body, kernel_proto] = self.buildOutlineFunction(clause_dict['REDUCTION'] + clause_dict['SHARED'], function_body)
        kernel_call = self.buildOutlineCall(clause_dict['REDUCTION'] + clause_dict['SHARED'], function_body)
        return [kernel_proto, kernel_call]


    
    def buildKernel(self, clause_dict, loop, ast):
        """ Build CUDA Kernel code """
        # ------------------- # TODO: Move this to some kind of template function
        # --------------------------------------------- def decls_to_param(elem):
        # ------------------------- if isinstance(elem.type, c_ast.ArrayDecl):
        # -------------------------------- return "*" + elem.name + "_cu"
        # --------------------------------------------------- return elem.name     
        # Get information from clauses
        reduction_params = clause_dict['REDUCTION']
        # TODO: Implement nowait
        # nowait = clause_dict.has_key('NOWAIT')
        # Check if nested for
        # TODO: Implement COLLAPSE instead of Nested For
        # collapse = clause_dict.has_key('COLLAPSE')
        # Private declarations come from the parent parallel construct
        private_params = clause_dict['PRIVATE']
        shared_params = clause_dict['SHARED']
        reduction_vars = get_template_array(reduction_params, ast)
        # Retrieve list of shared vars and build the array to template parsing
        loop_list = [loop.init.lvalue.name]
        private_vars = get_template_array([var for var in private_params if not var.name in loop_list], ast)
        loop_vars = get_template_array([var for var in private_params if var.name in loop_list], ast)
        if len(loop_vars) == 0:
            raise AbortMutationException(" No loop vars found: check your private clause\n")
        shared_vars = get_template_array(shared_params, ast,) 
        clause_vars = []
        clause_vars.extend(shared_vars)
        clause_vars.extend(private_vars)
       
        typedef_list = [str(decl) for decl in GlobalTypedefFilter().iterate(ast)]

        # Required type definitions
#        typedef_list = get_typedefs_to_template(clause_vars,ast)
        
        
        template_code = """
         %for line in typedefs:
                ${line}
         %endfor
        """
        
        template_code = BackendTemplateParser(self._backend_name, template_code).render(**{'typedefs' : typedef_list,})
        self.source_storage.append('llcomp_cuda', SourceStorage.C_HEADER, template_code)
        # Write functions used inside kernel to device
        ImplementOnCudaMutator(source_storage = self.source_storage, file_ast = ast, include_file = self._kernel_name).apply_all(loop.stmt)
        # TODO: This is incorrect, we should write a subtree instead of a bare string...
        for elem in reduction_params:
            IDNameMutator(old = c_ast.ID(name = elem.name, parent = elem.parent), new = c_ast.ID(name = 'reduction_cu_' + str(elem.name) + '[idx]', parent = elem.parent)).apply_all(loop.stmt)

        # OpenMP shared vars are parameters of the kernel function
        for elem in shared_params:
            # Replace the name of the declaration in the kernel code. @IndentOk
            if isinstance(elem.type, c_ast.ArrayDecl) or isinstance(elem.type, c_ast.Struct):
                mut = IDNameMutator(old = c_ast.ID(elem.name), new = c_ast.ID(elem.name + '_cu'))
                mut.apply_all(loop.stmt)

        template_code = open(config.WORKDIR + '/Backends/Cuda/Templates/kernel.tmpl', 'r').read()
        
#        tree = self.parse_snippet(template_code, {'kernelName' : self._kernel_name, 'reduction_vars' : reduction_vars, 'shared_vars' : shared_vars, 'typedefs' : typedef_list, 'private_vars' : private_vars, 'loop_vars' : loop_vars, 'loop' : loop} , name = 'KernelBuild', show = False)
        template_code = BackendTemplateParser(self._backend_name, template_code).render(**{'kernelName' : self._kernel_name, 'reduction_vars' : reduction_vars, 'shared_vars' : shared_vars, 'typedefs' : typedef_list, 'private_vars' : private_vars, 'loop_vars' : loop_vars, 'loop' : loop})
        self.source_storage.append(self._kernel_name, SourceStorage.CUDA_SOURCE, template_code)
        # All kernel build methods return a FileAST
        return None # c_ast.FileAST(ext = [tree.ext[-1]])


class FuncToDeviceMutator(AbstractMutator):
    """  Replace the definition of a FuncCall with a CUDAKernel with type __device__
     
    """
    def __init__(self, source_storage, func_call, file_ast, include_file):
        self.func_call = func_call
        self.source_storage = source_storage
        # Functions already defined on CUDA
        self.ignore_names = ['sqrt', 'min', 'max', 'log', 'fabs']
        self.file_ast = file_ast
        self.include_file = include_file
 
    def filter(self, subtree):
        """ Find the declaration of a function 
        """
        id_node = None
        if self.func_call.name.name in self.ignore_names:
            raise IgnoreMutationException('This function is already implemented on CUDA')
        try:
            decl = decl_of_id(self.func_call.name, self.file_ast)
            if hasattr(decl, 'storage') and 'extern' in decl.storage:
                print "********************"
                print dir(decl)
                decl.show()
                print "********************"
                print " *** Cannot use external declarations inside kernel *** "
                print " Decl : " + decl.name
                raise FilterError("Cannot use external declarations inside kernel")
            af = FuncDeclOfNameFilter(name = self.func_call.name)
            id_node = af.apply(self.file_ast)
        except NodeNotFound:
            print " *** Node not found *** "
            return None

        # Check if it is a first level declaration
        if isinstance(id_node.parent.parent, c_ast.FileAST):
            for elem in id_node.parent.parent.ext:
                if isinstance(elem, c_ast.FuncDef) and elem.decl.name == self.func_call.name:
                    return elem
            raise NodeNotFound(id_node)
        else:
            return id_node.parent.parent

    def mutatorFunction(self, ast):
        template_code = """
            __device__  ${decl}
        """
        template_code = BackendTemplateParser("FuncToDevice", template_code).render(**{'decl' : ast})
        self.source_storage.append('llcomp_cuda', SourceStorage.C_HEADER, template_code)
        return ast



class ImplementOnCudaMutator(AbstractMutator):
    """ Looks for declarations inside kernel and then looks for they definition
     
    """
    def __init__(self, source_storage, file_ast, include_file):
        self.source_storage = source_storage
        self.file_ast = file_ast
        self.include_file = include_file


    def filter_iterator(self, subtree):
        return FuncCallFilter().iterate(subtree)

    def mutatorFunction(self, subtree):
        return FuncToDeviceMutator(source_storage = self.source_storage, 
                                   func_call = subtree, file_ast = self.file_ast, 
                                   include_file = self.include_file).apply(subtree)


