# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

# from Frontend.C import c_ast

from Frontend.Shortcuts import getCurrentLanguageAst
from Backends.C99.Visitors.LlcFilters import AccScopeReverse
from Backends.Frangollo.Templates.FrangolloVarNode import FrangolloVarNode 

c_ast = getCurrentLanguageAst

from Backends.Common.Visitors.GenericVisitors import *

from Tools.Tree import InsertTool, NodeNotFound, ReplaceTool, RemoveTool, getContainerAttribute
from Tools.Declarations import type_of_id, decl_of_id
from Tools import SourceStorage

from Tools.Debug import DotDebugTool
# from Frontend.Parse import parse_source
from Backends.Common.Mutators.AstSupport import IDNameMutator
from Backends.Common.Mutators.AbstractMutator import IgnoreMutationException, \
                                    AbstractMutator, AbortMutationException
from Backends.Frangollo.Visitors.CM_Visitors import PragmaAccLoopFilter, AccLoopFilter

import config

from Tools import SourceStorage

from MiddleEnd.Functions.Mutators import Outliner
from MiddleEnd.Loop.Analysis import AccLoopParametrize, ParametrizeLoopTool

from Frontend.SymbolTable import updateSymbolTableFromNode
from Frontend.FrontendFactory import FrontendFactory
from Frontend.InternalRepr import AstToIR
from Frontend.C99.C99InternalRepr import C99AstToIR

from Backends.Common.TemplateEngine.TemplateParser import TemplateParser, BackendTemplateParser
from Backends.C99.Writers.AccWriter import AccWriter

from Backends.Frangollo.Runner import frangolloParser

class FrangolloMutatorError(Exception):
    def __init__(self, description):
        self.description = description

    def __str__(self):
        return "FrangolloMutatorError :: " + self.description


class AbstractFrangolloMutator(AbstractMutator):
    """ Common methods to work with Frangollo

    """
    def __init__(self, clauses = None, kernel_name = 'loopKernel', kernel_prefix = '', source_storage = None, fname = "unknown", **kwargs):
        if clauses == None:
            clauses = {}
        self.kernel_name = kernel_name
        self.kernel_prefix = "FM_" + kernel_prefix
        self._func_def = None
        self._parallel = None
        self._clauses = clauses
        self.device = "Frangollo"
        self._backend_name = 'Frangollo'
	self.fname = fname
        self.source_storage = source_storage
        super(AbstractFrangolloMutator, self, ).__init__(backend_name = self._backend_name)
 
    def lookParentWithAttr(self, parent_node, prev, attr):
        # Can be prev == parent_node ? If so uncomment this: 
        #prev = parent_node
        #parent_node = parent_node.parent
        while not hasattr(parent_node, attr):
            if isinstance(parent_node, c_ast.FileAST):
                raise MutationError
            prev = parent_node
            parent_node = parent_node.parent
        
        if prev == None:
            raise MutationError
        return prev, parent_node

    def parse_snippet(self, template_code, subs_dir, name, show = False, languages = None, writer = AccWriter):
        """ Transform a template code snippet to the IR

            :param template_code: String with the source code
            :param subs_dir: Dictionary with substitutions
            :param name: Name of the template to be parsed (debugging)
            :param show: Show the template after variable substitution (debugging)
            :return: IR-1 representation of the snippet

        """
        subtree = None
        if subs_dir:
            template_code = BackendTemplateParser(self._backend_name, template_code).render(**subs_dir)
            if show:
                print " Template " + str(template_code)
        try:
            subtree_ast = frangolloParser.parse_reuse(template_code, name, )
            # Transform the C99 ast into the internal representation
            migrator = C99AstToIR(Writer = writer, ast = subtree_ast)
            migrator.annotate()
            subtree_IR = migrator.ast
        except IOError:
                print "Pipe Error"
                return None

        return subtree_IR;

       
    def  _get_scope_clause_dict(self, node,):
        """ Builds the OpenAcc scope dictionary 
        :param node: Node from which extract information
        :return: Dictionary of clauses
        """
        # from Tools.Debug import DotDebugTool
        # DotDebugTool(highlight = [node, node.parent, node.parent.parent]).apply(getFileNode(node))
        llcfilt = AccScopeReverse(rootNode = getFileNode(node))
        llcfilt.apply(node)
        return llcfilt.get_clauses()
    
    def extract_directionality(self, ast, clauses, dGraph = None):
        names = []
	# RR: Reduction variables are INOUT to force recovering the result at output
        CLAUSES_TO_DIRECTION = {"COPYIN":'IN', "COPYOUT":'OUT', "COPY":'INOUT', 
                                "CREATE":'NONE', "PRIVATE" : 'PRIVATE', 
				"FIRSTPRIVATE" : 'FIRSTPRIVATE',
                                "PCOPYIN":'PIN', "PCOPYOUT":'POUT', "PCOPY":'PINOUT',
				"PRESENT":'PINOUT',
                                "REDUCTION":'PINOUT', 
                                "UNKNOWN":''}
        DIRECTIONS_TO_FRAN = {"IN" : 'VAR_IN', "OUT" : 'VAR_OUT', 
                              "INOUT" : 'VAR_INOUT', 
                              "PIN" : 'VAR_IN', "POUT" : 'VAR_OUT', 
                              "PINOUT" : 'VAR_INOUT', 
                              "PRIVATE": 'VAR_PRIVATE', 
                              "FIRSTPRIVATE": 'VAR_IN', 
			      "NONE" : "VAR_NONE", "UNKNOWN" : '' }
        for clause in CLAUSES_TO_DIRECTION.keys():
	    # Set to true if a variable requires the injection of a forced registration
	    force_register = False
            if clauses.has_key(clause):
                direction = DIRECTIONS_TO_FRAN[CLAUSES_TO_DIRECTION[clause]]
                symbol_list = None
                if clause == 'REDUCTION' and type(clauses[clause]) == type(list()):
                    if len(clauses[clause]) > 0 and type(clauses[clause][0]) == type(tuple()):
                       # Reduction clauses are ( REDUCTION_OPERATION, [ SYMBOL_LIST ] ) , so we need the 1 element
                        symbol_list = clauses[clause][0][1]
                    else:
                        symbol_list = clauses[clause]
		    force_register = True
		elif clause == 'FIRSTPRIVATE':
		    force_register = True
                    symbol_list = clauses[clause]
                else:
                    symbol_list = clauses[clause]
                for symbol in symbol_list:
                    canChange = True
                    if dGraph:
                        try:
                            canChange = dGraph[symbol].isWrite
                        except KeyError:
                            # Defaults to can Change
                            canChange = True
                    for elem in names:
                        if  elem.name == symbol.name:
                            break
                    else:
#                            print "Is present? "+ str(clause[0:2] == "PC")
                            names.append(FrangolloVarNode(symbol=symbol, 
                                    ast=ast, 
                                    direction= direction,
                                    canChange = canChange,
                                    present = (clause[0:2] == "PC"),
				    force_register = force_register
                                    ))
                
        return names, ast

    def compute_dgraph(self, subtree):
        from MiddleEnd.DataAnalysis.LlcScope import DGraph, ArrayIndexAnalysis
        dGraph = DGraph()
        st = getSymbolTableFromNode(subtree)
        aia = ArrayIndexAnalysis(dGraph, st) # Get index Variables
        aia.visit(subtree) 
        return dGraph

class FuncToDeviceMutator(AbstractMutator):
    """  Replace the definition of a FuncCall with a FrangolloKernel with type __device__
     
    """
    def __init__(self, source_storage, func_call, file_ast, include_file):
        self.func_call = func_call
        self.source_storage = source_storage
        # Functions already defined on Frangollo
        self.ignore_names = ['sqrt', 'min', 'max', 'log', 'fabs']
        self.file_ast = file_ast
        self.include_file = include_file
 
    def filter(self, subtree):
        """ Find the declaration of a function 
        """
        id_node = None
        if self.func_call.name.name in self.ignore_names:
            raise IgnoreMutationException('This function is already implemented on Frangollo')
        try:
            decl = decl_of_id(self.func_call.name, self.file_ast)
            if hasattr(decl, 'storage') and 'extern' in decl.storage:
                print "********************"
                print dir(decl)
                decl.show()
                print "********************"
                print " *** Cannot use external declarations inside kernel *** "
                print " Decl : " + decl.name
                raise FilterError("Cannot use external declarations inside kernel")
            af = FuncDeclOfNameFilter(name = self.func_call.name)
            id_node = af.apply(self.file_ast)
        except NodeNotFound:
            print " *** Node not found *** "
            return None

        # Check if it is a first level declaration
        if isinstance(id_node.parent.parent, c_ast.FileAST):
            for elem in id_node.parent.parent.ext:
                if isinstance(elem, c_ast.FuncDef) and elem.decl.name == self.func_call.name:
                    return elem
            raise NodeNotFound(id_node)
        else:
            return id_node.parent.parent

    def mutatorFunction(self, ast):
        template_code = """
            __device__  ${decl}
        """
        template_code = TemplateParser(template_code).render(**{'decl' : ast})
        self.source_storage.append('llcomp_Frangollo', SourceStorage.C_HEADER, template_code)
        return ast



class ImplementOnFrangolloMutator(AbstractMutator):
    """ Looks for declarations inside kernel and then looks for they definition
     
    """
    def __init__(self, source_storage, file_ast, include_file):
        self.source_storage = source_storage
        self.file_ast = file_ast
        self.include_file = include_file


    def filter_iterator(self, subtree):
        return FuncCallFilter().iterate(subtree)

    def mutatorFunction(self, subtree):
        return FuncToDeviceMutator(source_storage = self.source_storage, 
                                   func_call = subtree, file_ast = self.file_ast, 
                                   include_file = self.include_file).apply(subtree)

class AccParallelLoopMutator(AbstractMutator):
    """ 
    """
    def __init__(self, om):
        self.level = 0
        self.maxLevel = 0
        self.reduction_parameters = []
        self._om = om 
        #self.__noi__ = []
        super(AccParallelLoopMutator, self).__init__()

    def filter(self, ast):
        return PragmaAccLoopFilter().apply(ast)
        
    def fast_filter(self, ast):
        return PragmaAccLoopFilter().dfs_iter(ast)
        
    def filter_iterator(self, ast):
        return PragmaAccLoopFilter().iterate(ast)
        
    def mutatorFunction(self, ast):
        #import pdb
        #pdb.set_trace()
        # Get loop tagged
        accloop = AccLoopFilter().apply(ast)      
        #~ assert  accloop.stmt == c_ast.For
	updateSymbolTableFromNode(accloop)
        forloop = accloop.stmt
        
        # Update nesting level
        self.level += 1
        
        # Parametrize Acc loop
        parametrizer = AccLoopParametrize()
        loop_parameters = parametrizer.apply(forloop,)        
        print " Acc parallel loop", loop_parameters['loop_variable'], "with nesting level", self.level
        
        # Parameters for reductions
        if self.level <= 2:     # reduction in 3D loops not yet supported
            _merge_reduction_caluses_with(self.reduction_parameters, loop_parameters['reduction_vars'])
        
        # Get nested loop index variable
        if self.level == 1:
            #_index_var = '__idx__'
            _dim = 0
        elif self.level == 2:
            if loop_parameters['reduction_vars']:
                print " Reduction found in nested loop", loop_parameters['loop_variable'], ". Loop serialized "
                ReplaceTool(new_node = ast.stmt.stmt, old_node=ast               
                      ).apply(ast.parent, getContainerAttribute(ast))   
                return
            else:
                #_index_var = '__ydx__'
                _dim = 1
        else:
            print " 3D loops not yet supported"
            # Remove Pragma acc loop without mutating
            ReplaceTool(new_node = ast.stmt.stmt, old_node=ast               
                      ).apply(ast.parent, getContainerAttribute(ast))
            return 
        self.maxLevel = self.level
        
        #self.__noi__ += [loop_parameters['number_of_iterations']];
        
        # Code inside loop
        system_parameters = {}
        system_parameters['code'] = forloop.stmt
        #system_parameters['index_var'] = _index_var
        system_parameters['dim'] = _dim

        template = """
            /* TODO: Fix this */
            <%!
                from Backends.Common.TemplateEngine.Functions import var_to_declaration, id_to_var_node
                from Backends.C99.Writers.AccWriter import AccWriter
                from Frontend.InternalRepr import AstToIR
                from Tools.Declarations import type_of_id
                from Tools.Tree import getFileNode
            %>
            #define __type_loop_var ${type_of_id(l_p['loop_variable'], getFileNode(l_p['loop_variable']))}
            
            %for var in typedef:
                ${var};
            %endfor
            
            %for definition in getFileNode(l_p['loop_variable']).children():
            %if not (hasattr(definition, 'decl') and (definition.decl.name == 'foo' or definition.decl.name == 'fake')):
                ${definition};
            %else:
                /* NOTHING Decl */
            %endif            
            %endfor
         
         void foo {
            
            /* USEFULL BLOCK */
            {
                /* We compute number of tile blocks */
                __type_loop_var _stride_block_${l_p['loop_variable']} = ((${l_p['last_iteration']} - 1) / __yacf_get_num_groups(${s_p['dim']})) + 1;
                __type_loop_var _stride_thread_${l_p['loop_variable']} = (((${l_p['last_iteration']} - 1) / __yacf_get_num_groups(${s_p['dim']})) / __yacf_get_local_size(${s_p['dim']})) + 1;
        
                /* We compute index loop tiling variables */
                __type_loop_var _block_${l_p['loop_variable']} = ${l_p['init_iteration']} + __yacf_get_group_id(${s_p['dim']})  * _stride_block_${l_p['loop_variable']};
                __type_loop_var _thread_${l_p['loop_variable']} = _block_${l_p['loop_variable']} + __yacf_get_local_id(${s_p['dim']}) * _stride_thread_${l_p['loop_variable']};
            
                /* We need to check if loop_variable is inside limits (in all those cases we have more threads than iterations) */              
                __type_loop_var _last_block_${l_p['loop_variable']} =
                    (((_block_${l_p['loop_variable']} + _stride_block_${l_p['loop_variable']}) < ${l_p['last_iteration']}) ? _block_${l_p['loop_variable']} + _stride_block_${l_p['loop_variable']} : ${l_p['last_iteration']});
                  
                __type_loop_var _last_thread_${l_p['loop_variable']} =
                    (((_thread_${l_p['loop_variable']} + _stride_thread_${l_p['loop_variable']}) < _last_block_${l_p['loop_variable']}) ? _thread_${l_p['loop_variable']} + _stride_thread_${l_p['loop_variable']} : _last_block_${l_p['loop_variable']});
                
                /* Finally we have got loop iterating over tiled iterations in each thread */
                for (__type_loop_var ${l_p['loop_variable']} = _thread_${l_p['loop_variable']}; ${l_p['loop_variable']} < _last_thread_${l_p['loop_variable']}; ${l_p['loop_variable']}++)
                {
                  %if hasattr(s_p['code'], 'block_items'):
                  ${";\\n".join(str(elem) for elem in AstToIR(Writer=AccWriter, ast=s_p['code']).ast.block_items)};
                  %else:
                  ${AstToIR(Writer=AccWriter, ast=s_p['code']).ast};
                  %endif
                }
            }
         }
        """   
        from Backends.C99.Writers.AccWriter import AccWriter
        new_ast = self.parse_snippet(template, {'l_p' : loop_parameters,
                                                    's_p' : system_parameters,
                                                'typedef' : self._om._typedefList}, 
                                                name = 'AccParallelLoopMutator', show = False,
                                                writer = AccWriter)
        
        ReplaceTool(new_node = new_ast.ext[-1].body, old_node=ast               
                      ).apply(ast.parent, getContainerAttribute(ast))
        
        # Apply mutatorFunction to nested loops
        self.apply(new_ast)

        # Whenever the mutator returns from a recursive call, the value of the level variable is decremented 
        self.level -= 1
        
        return new_ast
        
def _merge_reduction_caluses_with(clauses_result, clauses_source):
    #import pdb
    #pdb.set_trace()
    # Constructs a set with reduction_vars name
    rvars = set()
    for op, _vars in clauses_result:
        for var in _vars:
            rvars.add(var.name)
    # Adds reduction var to clauses_result if not yet in it
    for op, _vars in clauses_source:
        for var in _vars:
            if var.name not in rvars:
                clauses_result += [(op, [var])]
