# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------


import copy
import config

from Backends.C99.Visitors.LlcFilters import AccKernelsFilter, AccLoopFilter,\
    AccScopeReverse, AccDataFilter

from Tools.Tree import InsertTool, NodeNotFound, ReplaceTool, getFileNode
from Frontend.C99 import c99_ast
from Frontend.Acc import acc_ast
from Tools.Declarations import decl_of_id

from Tools import SourceStorage
from Backends.Common.Mutators.AbstractMutator import AbstractMutator, MutatorException
from Backends.Frangollo.Templates.FrangolloVarNode import FrangolloVarNode
from Backends.Frangollo.Mutators.FM_LlcFor import FM_AccLoop
from Backends.Frangollo.Mutators.FM_PrivDecl import FM_PrivDecl

from MiddleEnd.DataAnalysis.LlcScope import DNode,DGraph,BlockDepGraphBuilder

from Frontend.SymbolTable import getSymbolTableFromNode, updateSymbolTableFromNode
from Backends.Frangollo.Mutators.Common import AbstractFrangolloMutator

from Backends.Frangollo.Mutators.FM_AccHostData import FM_AccHostData

from Backends.Frangollo.Mutators.FM_AccUpdate import FM_AccUpdate

from Frontend.InternalRepr import AstToIR


# TODO: Move this to a external module
def _merge_dict_per_key(destination, source):
        """ Merge two clause dicts per key value. 
            Sample:
            .. ipython:
            
            [0] a = { 'A' : [0,1], 'B' : [2,], 'C' : []}
            [1] b = { 'B' : [3,] , 'C' : 3 }
            [2] c = _merge_dict_per_key(a, b)
            [3] c = {'A' : [0,1], 'B' : [2,3], 'C' : 3}
            
        """
        source_cpy = copy.copy(source)
        for elem in destination:
            if source_cpy.has_key(elem):
                if type(destination[elem]) == type([]) and (len(destination[elem]) > 0):
                    destination[elem] += source[elem]
                else:
                    if type(source[elem]) == type([]) and (len(source[elem]) > 0):
                        destination[elem] = source[elem]
                    elif type(source[elem]) != type([]):
                        destination[elem] = source[elem]
                    else:
                        # Keep destination
                        pass
                del source_cpy[elem]
            
        destination.update(source_cpy)

        return destination        



class FM_AccRegion(AbstractFrangolloMutator):
    def __init__(self, *args, **kwargs):
        self._parentRegion = None
        self._source_storage = kwargs['source_storage']
        self._backend_name = "Frangollo"
        self._kernelList = []  # List of device kernels contained within this Context
        self._isDataDirective = False
        self._isParallelDirective = False
        self._isKernelsDirective = False
        self._contextCreationTemplate =  """
        #include "openacc.h"
        
        // Context substitution
        int main() {
         int num_kernels = ${len(region['kernelList'])} + 1;
         tkernel_list *list = (tkernel_list *) malloc (num_kernels*sizeof(tkernel_list));
         strcpy(list[(--num_kernels)].kernel_name, "END");
         %for kdata in region['kernelList']:
             strcpy(list[(--num_kernels)].kernel_name, "${kdata["name"]}");
             // list[num_kernels].numFlop = ${kdata["flops"]};
             // list[num_kernels].numMemoryAccess = ${kdata["memAccess"]};
         %endfor
         int regionId = FRG__createContext("${region['name']}", list);
         FRG__setActiveContext(regionId);
         %for var in region['vars']:
             %if var.frangolloType == "VAR_SCALAR" and var.direction != "VAR_PRIVATE":
                {
                var_layout_t var_layout, acc_layout;
                var_layout.x = ${var.size};
                var_layout.y = 0;
                var_layout.type = ${var.frangolloType};
                FRG__registerVar(&(${var.name}), regionId, ${var.present} | 
                                    (${var.direction} * 0x10) | 
                                    (${var.canChange} * 0x100),
                                    var_layout, &acc_layout);
                }

             %elif var.frangolloType == "VAR_ARRAY2D":
                {
                int __acc__lda_${var.name} = ${var._matrixInfo.lda};
                var_layout_t var_layout, acc_layout;
                var_layout.x = ${var._matrixInfo.height};
                var_layout.y = ${var._matrixInfo.width};
                var_layout.elemSize = sizeof(${var._symbol.btype});
                var_layout.lda = ${var._matrixInfo.lda};
                var_layout.type = ${var.frangolloType};
                FRG__registerVar((${var.name}), regionId, ${var.present} | 
                                    (${var.direction} * 0x10) | 
                                    (${var.canChange} * 0x100),
                                    var_layout, &acc_layout);
               
                }
             %elif var.direction != "VAR_PRIVATE":
             {
                     var_layout_t var_layout, acc_layout;
                     var_layout.x = ${var.size};
                     var_layout.y = 0;
                     var_layout.lda = ${var.size};
                     var_layout.type = ${var.frangolloType};
                %if var._symbol.extra['startAddr']:
							//  sizeof(${var._symbol.btype})
                             FRG__registerVar((${var.name}) + (${var._symbol.extra['startAddr']} * sizeof(${var.type.type})), 
                                        regionId, ${var.present} | 
                                          (${var.direction} * 0x10) | 
                                          (${var.canChange} * 0x100),
                                        var_layout, &acc_layout);
                %else:
                     FRG__registerVar((${var.name}), regionId, ${var.present} | 
                                     (${var.direction} * 0x10) | 
                                     (${var.canChange} * 0x100),
                                     var_layout, &acc_layout);
                %endif
             }
             %endif
         %endfor
		
         {
             placeholder();
         }
         {
                 FRG__destroyContext(regionId);
                 %if not region['async']:
                    FRG__barrier();
                 %endif
         }
        }
        """
        if kwargs.has_key('xmlm'):
#            print " Has key, value: " + str(kwargs['xmlm'])
            self._xmlm = kwargs['xmlm']
        else:
            self._xmlm = None
        super(FM_AccRegion, self).__init__(*args, **kwargs)
    
    def getExtractedKernelList(self,):
        return self._kernelList




    def mutatorFunction(self, ast, accKernels_node):
        """ Frangollo mutator, creates a context and prepares the associate kernels
        """
        if (not self._isDataDirective and not self._isKernelsDirective and not self._isParallelDirective):
            import pdb
            pdb.set_trace()
            raise MutatorException(" Trying to create a context with an unidentified region, abort \n")
        
        
        ##################### Statement for Frangollo region
        self._st = getSymbolTableFromNode(accKernels_node)
        accKernels_code = self._contextCreationTemplate

        clauses = {}
        mixed_construct = False
        if isinstance(accKernels_node.stmt, c99_ast.Compound):
            if (hasattr(accKernels_node.stmt.block_items[0],'clauses') \
              and isinstance(accKernels_node.stmt.block_items[0], acc_ast.AccLoop)):
                  # It is a Kernel/Parallel + loop
                  mixed_construct = True
                  # print "... It is a Kernel/Parallel + loop"
                  clauses = self._get_scope_clause_dict(accKernels_node.stmt.block_items[0].stmt);
            elif (hasattr(accKernels_node,'clauses') \
              and isinstance(accKernels_node, acc_ast.AccParallel)):
                  # It is a Parallel
                  clauses = self._get_scope_clause_dict(accKernels_node.stmt);                 
        if config.IGNORE_SCOPE:
            print " ----> Warning: Ignoring scope for directives "
            from Backends.C99.Visitors.LlcFilters import AccScopeReverse as ASR
            clauses = ASR(rootNode = getFileNode(accKernels_node))._get_dict_from_clauses(accKernels_node.clauses)
        # import pdb
        # pdb.set_trace()
        _merge_dict_per_key(clauses, self._get_scope_clause_dict(accKernels_node.stmt))
        

        region = {}
        region['name'] = self.fname + "_"
        if clauses.has_key('NAME'):
            region['name'] += clauses['NAME'] or "random"
            print "[c2frangollo] Generating code for with name %s "%(region['name'])
        else:
            region['name'] += "random"
            print "[c2frangollo] Generating code for region around line %d "%(accKernels_node.parent.coord.line)


        if clauses.has_key('ASYNC'):
            region['async'] = clauses['ASYNC']
        else:
            region['async'] = None

        # Extract directionality of context code BEFORE mutating internals
        dGraph = None
        
        if not self._isDataDirective:
            # Extract dependency
            dGraph = self.compute_dgraph(accKernels_node)
        names, ast = self.extract_directionality(ast, clauses, dGraph)
        region['vars'] = names
        # Check that we have directionality information for all variables
        external_vars = [ _id.name for _id in self._getUndeclVars(accKernels_node) ]
        for var in names:
            if var.name in external_vars:
               external_vars.remove(var.name)
        if len(external_vars) > 0:
            print "Warning: Following variables were not declared in any clause"
            for elem in external_vars:
                print elem + "," ,
            print " at region " + (region['name'])
            # raise MutatorException 

        ######### TODO: This is a bad design practise (TM) , since it relies on 
        #               on a set of if/else to identify actions. This *must* be refactored
        #               into something that is handled by the children dynamically
        if not self._isDataDirective and (self._isKernelsDirective):
            #fm_privDecl = FM_PrivDecl(clauses = clauses, source_storage = self._source_storage)
            #fm_privDecl.apply_all(accKernels_node, getFileNode(accKernels_node))
            # 3. Generate Kernel codes from For constructs
            fm_AccLoop_mutator = FM_AccLoop(clauses = clauses, 
                                    kernel_name = region['name'] + "_loopKernel", 
                                    source_storage = self._source_storage, xmlm = self._xmlm, 
				    fname = self.fname, symbol_table = self._st)
            fm_AccLoop_mutator.apply_all(accKernels_node, getFileNode(ast))
            self._kernelList += fm_AccLoop_mutator.getExtractedKernelList()
        elif self._isParallelDirective:
            # Replace the statements by a kernel call
            # self._kernelList += self._generateParallel(accKernels_node, getFileNode(ast), region)
            self.build_basic_kernel(clauses, region, accKernels_node.stmt, True, True, dGraph)
            # self._kernelList += kl

        # Host data construct inside a Region
        fm_AccHostData = FM_AccHostData(clauses = clauses, source_storage = self._source_storage, xmlm = self._xmlm)
        fm_AccHostData.apply_all(accKernels_node, getFileNode(ast))
        # Update construct inside a Region, may appear outside??
        fm_AccUpdate = FM_AccUpdate(clauses = clauses, source_storage = self._source_storage, xmlm = self._xmlm)
        fm_AccUpdate.apply_all(accKernels_node, getFileNode(ast))


        region['kernelList'] = self._kernelList

        llcContext_subtree = self.parse_snippet(accKernels_code, 
                                             {'region' : region}, 
                                             name = 'General Context', 
                                              show = False).ext[-1].body

                                              
        next_accKernels_stmt = accKernels_node.stmt
       
        # Skips all consecutive pragmas
        while isinstance(next_accKernels_stmt, c99_ast.Pragma) and \
                hasattr(next_accKernels_stmt, 'stmt'):
            next_accKernels_stmt = next_accKernels_stmt.stmt
        
        # llcContext_subtree.block_items[-2] = Context contents
        ReplaceTool(new_node = next_accKernels_stmt, 
                    old_node = llcContext_subtree.block_items[-2]
                   ).apply(llcContext_subtree, 'block_items')
        
        
        ##################### pragma region replacement
        # Replace the entire pragma by a CompoundStatement with all the new statements
        # Note: The parent of Parallel is always a Pragma node
	try:
        	ReplaceTool(new_node = llcContext_subtree, 
                            old_node = accKernels_node.parent
                        ).apply(accKernels_node.parent.parent, 'block_items')
	except ValueError as ve:
		import pdb
		pdb.set_trace()
		#ReplaceTool(new_node = llcContext_subtree, 
                #            old_node = accKernels_node.parent
                #        ).apply(accKernels_node.parent.parent, 'block_items')

		print MutatorException("Cannot replace AccKernel node ")

    def _insertConcylePoints(self, llcContext_subtree):
        """ Insert memory concyle points inside the context code.
            Using a Statement Dependency Graph, we extract dependency information
            from the llcContext code in order to stablish with variables are required
            to be updated before a particular statement is run. Due to the fact that
            all variables are available to all devices, it is important that if the
            original code uses any variable, its position must be updated.

            :param llcContext_subtree: IR-1 with the statements of the Context

        """
        # Build Stmt dependency graph
        dGraph = DGraph()
        bdgb = BlockDepGraphBuilder(dGraph, self._st)
        bdgb.visit(llcContext_subtree)
        [predStmt, requiredVars] = bdgb.computeDependencyMatrix()
        
        def concyleCallFactory(placeholder, originalStmt):
               concyle_subtree = c99_ast.Compound(block_items = [],parent = None)
               concyleCall_subtree = c99_ast.FuncCall(name = c99_ast.ID(name = 'concyle'), args = [], parent = concyle_subtree)
               concyleCall_subtree.args = c99_ast.ExprList(exprs = [c99_ast.ID(name = 'regionId', parent = concyleCall_subtree), c99_ast.UnaryOp(op = '&', \
                                            expr = c99_ast.ID(name = placeholder), \
                                            parent = concyleCall_subtree)])
               concyle_subtree.block_items = [originalStmt, concyleCall_subtree]
               return concyle_subtree

        # Insert concyle stmts
        numberOfInsertions = 0
        pos = 0
        for stmt in requiredVars:
        #    if not type(requiredVars) == type([]):
        #        print "requiredVars  Not a list"
            for var in stmt:
 #               print " Insert concyle of " + str(var.name) + " before S" + str(pos) + " --" + str([s.name for s in stmt]) + "--" + str([[str(i) + ":" + str(i) for i in r] for r in requiredVars]) 
                newStmt = concyleCallFactory(var.name, llcContext_subtree.block_items[pos - 1])
                llcContext_subtree.block_items[pos - 1] = newStmt
                setattr(newStmt, 'parent', llcContext_subtree)
            pos = pos + 1


class FM_AccKernels(FM_AccRegion):
   
    def __init__(self, *args, **kwargs):
        super(FM_AccKernels, self).__init__(*args, **kwargs)
        self._isDataDirective = False
        self._isParallelDirective = False
        self._isKernelsDirective = True
    
    def filter(self, ast):
        """ Filter definition
            Returns the first node matching with the filter
        
        """
        
        f = AccKernelsFilter()
        node = f.apply(ast)
        self._parentRegion = node
        self._isKernelsDirective = True
        self._isParallelDirective = False
        self._isDataDirective = False
        return node

    def apply_all(self, ast):
        """ Apply mutation to all matches """
        self.ast = ast
        f = AccKernelsFilter()
        final_kernel_list = []
        try:
            for elem in f.iterate(ast):
                self._kernelList = []
                self.mutatorFunction(ast, elem)
                final_kernel_list += self._kernelList
        except NodeNotFound as nf:
            print str(nf)
        except StopIteration:
            self._kernel_list = final_kernelList
            return self._parentRegion

        self._kernelList = final_kernel_list
        return ast




class FM_AccData(FM_AccRegion):
     
    def __init__(self, *args, **kwargs):
        super(FM_AccData, self).__init__(*args, **kwargs)
        self._isDataDirective = True
        self._isParallelDirective = False
        self._isKernelsDirective = False

    def filter(self, ast):
        """ Filter definition
            Returns the first node matching with the filter
        
        """
        f = AccDataFilter()
        node = f.apply(ast)
        self._parentRegion = node
        self._isDataDirective = True
        return node

    def apply_all(self, ast):
        """ Apply mutation to all matches """
        self.ast = ast
        print " Filter called "
        f = AccDataFilter()
        self._isDataDirective = True
        try:
            for elem in f.iterate(ast):
                self.mutatorFunction(ast, elem)
        except NodeNotFound as nf:
            print str(nf)
        except StopIteration:
            return self._parentRegion
        return ast


   



