# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

from Tools import SourceStorage
import config
from Backends.Cuda.Files.FileTypes import Cuda_FileType
from Backends.C.Files.FileTypes import C_FileType, H_FileType
from Backends.Common.Files.FileTypes import Makefile_FileType, MK_FileType
from Backends.Common.TemplateEngine.TemplateParser import TemplateParser
import os
#### Frangollo makefile template

if config.nv_opts:
    nv_additional_opts = config.nv_opts
else:
    nv_additional_opts = ""

frangollo_make_template = open(config.WORKDIR+'/Backends/Frangollo/Templates/makefile.tmpl', 'r').read()

from Frontend.FrontendFactory import FrontendFactory

# frangolloParser = None

class FrangolloParserSingleton(FrontendFactory):
   # To generate Frangollo Code, we only need C99 and Acc parsers
   def __init__(self, includes = []):
      self._languages = ['C99','Acc',]
      self._prepro_class = self.getPreproFromLanguageList(['C99',], includes)
      self._lexer_class = self.getLexerFromLanguageList(self._languages)
      self._parser_class = self.getParserFromLanguageList(self._languages,)
      self._prepro = self._prepro_class()
      self._parser = self._parser_class(lex_optimize = config.LEX_OPTIMIZE, yacc_optimize = config.YACC_OPTIMIZE, lexer_class = self._lexer_class, yacc_debug=config.YACC_DEBUG)


frangolloParser = FrangolloParserSingleton()


class FrangolloTransformer:
    """ Apply the Frangollo Transformers to a given AST
    
            Create a SourceStorage object to save all files created
            by translations.
            A separate file is used for each kernel.
    """
    
    @staticmethod
    def apply(ast, fname, output = ".", original_file = None, xmlm = None, other_files = None):
        """ Run the OpenAcc to CUDA Translation
            
            From an ast, it produces a project directory with
                the translated sources.
        
        """
        from Backends.Frangollo.Mutators.FM_AccRegions import FM_AccKernels, FM_AccData
        from Backends.Frangollo.Mutators.FM_AccWait import FM_AccWait
        from Backends.Frangollo.Mutators.FM_AccParallel import FM_AccParallel
        #=======================================================================
        # _source_storage = SourceStorage.Storage('/', output + "/Frangollo/")
        # # Create header and support files for the Frangollo Target
        # _source_storage.addFile('llcomp_cuda', H_FileType())
        # _source_storage.append('llcomp_cuda', H_FileType(), template_header)
        #=======================================================================
	# print " Other files " + str(other_files)
        # Ensure is a valid fname
        clean_fname = "_" + fname
        _source_storage = SourceStorage.Storage('', output + "/Frangollo/")
        # Run the mutators
        # -> LlcContext mutator will find the rest of llc constructs inside a particular context
        fm_accParallel_mutator = FM_AccParallel(source_storage = _source_storage, xmlm = xmlm, fname = clean_fname)
        fm_accParallel_mutator.apply_all(ast)
        
        fm_accKernels_mutator = FM_AccKernels(source_storage = _source_storage, xmlm = xmlm, fname = clean_fname)
        fm_accKernels_mutator.apply_all(ast)
        fm_accData_mutator = FM_AccData(source_storage = _source_storage, xmlm = xmlm, fname = clean_fname)
        fm_accData_mutator.apply_all(ast)
        fm_accWait_mutator = FM_AccWait(source_storage = _source_storage, xmlm = xmlm, fname = clean_fname)
        fm_accWait_mutator.apply_all(ast)

        _source_storage.addFile(fname, C_FileType())
        if config.INTERNAL_HACKS:
            _source_storage.append(fname, C_FileType(), config.DEFAULT_INCLUDES)
        #from Frontend.C99.c99_prepro import C99Prepro
        #print " Includes: " + C99Prepro.includes
        _source_storage.append(fname, C_FileType(), '#include "openacc.h"\n')
        _source_storage.append(fname, C_FileType(), str(ast))

        kernel_list = fm_accKernels_mutator.getExtractedKernelList()        
        kernel_list.extend(fm_accParallel_mutator.getExtractedKernelList())
        
        subs_dir = {'kernel_list' : kernel_list,
                    'original_filename' : fname,
		              'other_files' : other_files,
                   }
                    
        frangollo_make_content = TemplateParser(frangollo_make_template).render(**subs_dir)

        _source_storage.addFile('Makefile', Makefile_FileType())
        _source_storage.append('Makefile', Makefile_FileType(), frangollo_make_content)

        _source_storage.addFile('makefile', MK_FileType()) 
        _source_storage.append('makefile', MK_FileType(), open(config.MK_FILE).read())
              
        # Storage write files when deleted
        del _source_storage
        
        return True
    
