// Simplified version of openacc include to ease parsing

typedef int size_t;

typedef struct _tparams{
   void * _base_ptr;
   int type;
   char* reduction;
   char* reductionType;
   char* identity;
   unsigned int canChange;
   struct _tparams * next;
} tparams;


#define MAX_KERNEL_NAME 255
#define MAX_KERNELS_PER_REGION 32

/**
* @brief Struct to store all the region's kernel names.
*/
typedef struct {
  char kernel_name[MAX_KERNEL_NAME];
  /* Information about kernel to help guessing 
        the proper thread/block combination */
  unsigned int numMemoryAccess;
  unsigned int numFlop;
} tkernel_list;


/* Variable memory layout */
typedef struct {
    size_t x, y;
    size_t lda;
    size_t pitch;
    size_t elemSize;
    int type;
   } var_layout_t;


#ifndef DEBUG
      #define NEW_KERNEL_PARAM(_varName, _basePtr, _type, _reductionOp, _reductionType, _identity, _canChange, _size) \
      { \
                  tparams  * __FRG__parameter_list_params = (tparams *) malloc(sizeof(tparams));\
                  if (!__FRG__parameter_list_head) {\
                          __FRG__parameter_list_head = __FRG__parameter_list_params;\
                  }\
                  if (__FRG__parameter_list_prev) {\
                          __FRG__parameter_list_prev->next = __FRG__parameter_list_params;\
                  }\
                  __FRG__parameter_list_prev = __FRG__parameter_list_params;\
                  __FRG__parameter_list_params->_base_ptr = _basePtr;\
                  __FRG__parameter_list_params->type = _type;\
                  __FRG__parameter_list_params->reduction = _reductionOp; /* 0x0 '+' '*'  */ \
                  __FRG__parameter_list_params->identity = _identity;\
                  __FRG__parameter_list_params->reductionType = _reductionType;\
                  __FRG__parameter_list_params->canChange = _canChange; \
                  __FRG__parameter_list_params->size = _size; \
                  __FRG__parameter_list_params->varName = NULL; \
                  __FRG__parameter_list_params->next = NULL; \
      }
#else
      #define NEW_KERNEL_PARAM(_varName, _basePtr, _type, _reductionOp, _reductionType, _identity, _canChange, _size) \
      { \
                  tparams  * __FRG__parameter_list_params = (tparams *) malloc(sizeof(tparams));\
                  if (!__FRG__parameter_list_head) {\
                          __FRG__parameter_list_head = __FRG__parameter_list_params;\
                  }\
                  if (__FRG__parameter_list_prev) {\
                          __FRG__parameter_list_prev->next = __FRG__parameter_list_params;\
                  }\
                  __FRG__parameter_list_prev = __FRG__parameter_list_params;\
                  __FRG__parameter_list_params->_base_ptr = _basePtr;\
                  __FRG__parameter_list_params->type = _type;\
                  __FRG__parameter_list_params->reduction = _reductionOp; /* 0x0 '+' '*'  */ \
                  __FRG__parameter_list_params->identity = _identity;\
                  __FRG__parameter_list_params->reductionType = _reductionType;\
                  __FRG__parameter_list_params->canChange = _canChange; \
                  __FRG__parameter_list_params->size = _size; \
                  __FRG__parameter_list_params->varName = NULL; \
                  __FRG__parameter_list_params->next = NULL; \
                  printf("(%s,%d) Adding parameter %s with ptr %p (RO? %c) \n", __FILE__, __LINE__, _varName, _basePtr, __FRG__parameter_list_params->canChange?'Y':'N'); \
      }
#endif


   /* Region identifier */
   typedef int ctxtId_t;



    /* Kind of variables */
    #define VAR_SCALAR 0
    #define VAR_ARRAY 1
    #define VAR_ARRAY2D 2
    /* Var. directionality */
  // Is used inside context but not required again outside
    #define VAR_IN 0x0         
// Is used outside after context but not init. before
    #define VAR_OUT 0x1         
  // Is required both pre-context and after-context
    #define VAR_INOUT 0x2       
// It exists only within the context
    #define VAR_PRIVATE 0x3      
    #define VAR_NONE VAR_PRIVATE
    #define VAR_RONLY 0x0
    #define VAR_RW 0x1
    #define VAR_WONLY 0x2
    #define FORCE_COPY_IN 0
    #define COPY_FIRST_ONLY 1


   /* acc_init( devicetype )
   Initialized the runtime system and sets the accelerator device
   type to use for this host thread.
   Usage example 
   acc_init(acc_device_type);
   */
   typedef  enum  {
      acc_device_none,
      acc_device_default,
      acc_device_host,
      acc_device_not_host,
      acc_device_nvidia,
      acc_device_generic_opencl
   } acc_device_t;

   void acc_init(acc_device_t device);

   // Returns the number of accelerator devices of the specified type.
   int acc_get_num_devices(acc_device_t d_type);


   // Sets the accelerator device type to use for this host thread.
   void acc_set_device_type(acc_device_t d_type);

   /* The acc_get_device_type routine tells the program what type of device will be used to
      run the next accelerator region, if one has been selected. This is useful when the
      implementation allows the program to be compiled to use more than one type of accelerator.*/
   acc_device_t acc_get_device_type ();

   // Sets the accelerator device number to use for this host thread..
   void acc_set_device_num(int devicenum, acc_device_t d_type);


   // Returns the accelerator device number that is being used by
   // this host thread.
   int acc_get_device_num(acc_device_t d_type);


   // Disconnects this host thread from the accelerator device.
   void acc_shutdown(acc_device_t d_type);


   /* In an OpenACC parallel or kernels region, this is used
      to take different execution paths depending on whether the
      program is running on an accelerator or on the host.*/
   void acc_on_device(acc_device_t d_type);


   /* Returns the address of memory allocated on the accelerator
      device.*/
   void * acc_malloc(size_t size);

   /* acc_free( void* )
   Frees memory allocated by acc_malloc.*/
   void acc_free(void * ptr);

   /* The acc_async_test routine tests for completion of all associated asynchronous activities.*/
   int acc_async_test( int );

   /* The acc_async_test_all routine waits for completion of all asynchronous activities. */
   int acc_async_test_all( );


   /* The acc_async_wait routine waits for completion of all associated asynchronous
      activities. */
   void acc_async_wait( int );


   /* The acc_async_wait_all routine waits for completion of all asynchronous activities. */
   void acc_async_wait_all( );


   /* Private rutine to add kernels to an existing context*/
   void __acc__addKernel(tkernel_list * kernel_name_list);


   /********* Frangollo interface *****************/
   void FRG__registerVar (void * var,    
            ctxtId_t cid,           
            int var_flags,          
            var_layout_t var_layout,
            var_layout_t * var_layout_p); 



   int FRG__submit_kernel_auto   (ctxtId_t regionID, char * kernelName, 
                          tparams * params, int size_x, int size_y, 
                          int size_z, int numFlops, int memAccess);
   
   int FRG__submit_kernel_manual (ctxtId_t regionID, char * kernelName, 
                          tparams * params,int blocksPerGrid_x, int threadsPerBlock_x,
                          int blocksPerGrid_y, int threadsPerBlock_y);

    ctxtId_t FRG__createContext(char * name,         // Name of the context
                        tkernel_list * kernel_name_list // List of potential kernels
                     );

    int FRG__destroyContext(ctxtId_t contextId      // Context identification
                       );

    int concyle(ctxtId_t contextId,            // Context identification
                void * varPtr             // Ptr to the original var to concyle
                );

    int FRG__setActiveContext(ctxtId_t contextId);
    ctxtId_t FRG__getActiveContext();


    int FRG__offloadKernel(ctxtId_t contextId, char * kernelName);
    
    void * FRG__getDevicePtr(ctxtId_t contextId, void * var);
    int FRG__update_in(ctxtId_t contextId, void * varPtr);
    int FRG__update_out(ctxtId_t contextId, void * varPtr);

    void FRG__printStats();
