# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

'''
Created on Jun 24, 2011

@author: rreyes
'''
from Backends.Common.TemplateEngine.TemplateParser import TemplateVarNode
from Tools.Tree import getFileNode
from Frontend.SymbolTable import getSymbolTableFromNode
from Backends.Common.TemplateEngine.Functions import is_complex_type

class GMACVarNode(TemplateVarNode):
    '''
    @author: Ruyman
    '''


    @property
    def direction(self):
        return self._directionality
    
    def __init__(self, var, ast, direction):
        """ Build the TemplateVarNode making a clausure
                with the modifier functions.
        """
        self._var = var
        self._ast = ast
        self._type = var.type
        self._type_func = lambda elem: elem.type
        self._name_func = lambda elem : elem.name
        
        # Get the symbol table from any node (they share the same FileAST)
        self._st = getSymbolTableFromNode(getFileNode(var))
        self._symbol = self._st.lookUpDecl(var)
        
        self._typedef = self._get_typedef_from_node(var, ast)
        self._structDecl = self._get_structdecl_from_node(var, ast)
        self._directionality = direction
        self._reductionOp = 0x0
    
    @property
    def frangolloType(self,):
        if is_complex_type(self._var):
            return "VAR_ARRAY"
        else:
            return "VAR_SCALAR"
 

    @property
    def reductionOp(self,):
        """ If a reduction operation is to be applied within this variable
                           this property specifies the reduction operation.
        """
        return self._reductionOp


    @reductionOp.setter
    def reductionOp(self, value):
        """ If a reduction operation is to be applied within this variable
                           this property specifies the reduction operation.
        """
        self._reductionOp = value


FRANGOLLO_VAR_DIRECTIONS = {"IN" : 'VAR_IN', "OUT" : 'VAR_OUT', "INOUT" : 'VAR_INOUT'}



def createRegionVarList(copyIn_clause, copyOut_clause, ast,):
    """ Convert an array list to a template array 
    
        :param var_list: List with Decl elements
        :param ast: Subtree
        :param func: Type of element from var_list which is added to the template array
        :param name_func: Function to get the name of each element (default elem.name)
        :param type_func: Function to get the type of each element (default elem.type)
        :return: List with TemplateVarNode elements
    
    """
    names = []
    onlyOut = copyOut_clause[:]

    for in_decl in copyIn_clause:
        isIn = True
        isOut = False
        for out_decl in copyOut_clause:
            if out_decl.name == in_decl.name:
                isOut = True
                onlyOut = [ elem for elem in onlyOut if elem.name != out_decl.name]
                break
        if isIn and isOut:
            direction = 'INOUT'
        elif isIn and not isOut:
            direction = 'IN'
        elif not isIn and isOut:
            direction = 'OUT'
        else:
            direction = 'INOUT'
           
        names.append(GMACVarNode(var = in_decl, ast = ast, direction = FRANGOLLO_VAR_DIRECTIONS[direction]))

    for out_decl in onlyOut:
        names.append(GMACVarNode(var = out_decl, ast = ast, direction = FRANGOLLO_VAR_DIRECTIONS['OUT']))


           
    return names 
 
def createKernelParameterList(copyIn_clause, copyOut_clause, kernel_parameters, reduction_clause, ast,):
    """ Convert an array list to a template array 
    
        :param var_list: List with Decl elements
        :param ast: Subtree
        :param func: Type of element from var_list which is added to the template array
        :param name_func: Function to get the name of each element (default elem.name)
        :param type_func: Function to get the type of each element (default elem.type)
        :return: List with TemplateVarNode elements
    
    """
    # Kernel parameters are : (shared + reduction - local_vars)
    names = []
#    names.append(GMACVarNode(var = elem, ast = ast, direction = FRANGOLLO_VAR_DIRECTIONS['INOUT']))
    for decl in kernel_parameters:
        isIn = False
        isOut = False
        for in_decl in copyIn_clause:
            if in_decl.name == decl.name:
                isIn = True
                break
        for out_decl in copyOut_clause:
            if out_decl.name == decl.name:
                isOut = True
                break
        if isIn and isOut:
            direction = 'INOUT'
        elif isIn and not isOut:
            direction = 'IN'
        elif not isIn and isOut:
            direction = 'OUT'
        else:
            direction = 'INOUT'
        names.append(GMACVarNode(var = decl, ast = ast, direction = FRANGOLLO_VAR_DIRECTIONS[direction]))

   
    # Add which reduction operation (if any) is going to be applied to the GMACVar
    for op,varlist in reduction_clause:
        for elem in varlist:
           for fvn in names:
                if elem.name == fvn.ptr.name:
                    fvn.reductionOp = "'" + op + "'"

    return names    
