# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

# -*- coding: utf-8 -*-

'''
 :author: Rayco Abad-Martín <rayco.abad@gmail.com>
 :author: Elena Marrero-Méndez <elemarez@gmail.com>
'''

from Backends.C.Files.FileTypes import H_FileType, C_FileType
from Backends.C.Visitors.OmpFilters import *
from Backends.MPI.Mutators.Common import AbstractMPIMutator, mpi_outline_loop_stmts
from Backends.MPI.Mutators.MPI_OmpFor import *
from Frontend.Shortcuts import getCurrentLanguageAst
from Frontend.SymbolTable import updateSymbolTableFromNode
from Tools import SourceStorage
from Tools.Tree import InsertTool, NodeNotFound, ReplaceTool, RemoveTool, getFileNode

c_ast = getCurrentLanguageAst()

class MPI_OmpParallelFor(AbstractMPIMutator):
    def __init__(self, file_name = 'loopParallelFor', *args, **kwargs):
        self.file_name = file_name
        super(MPI_OmpParallelFor, self).__init__(*args, **kwargs)

    def filter(self, ast):
        """ Filter definition
            @return the first node matching with the filter
        """
        # Build a visitor , matching the OmpParallelFor node of the AST
        f = OmpParallelForFilter(device = self.device)
        node = f.apply(ast)
        self._func_def = f.get_func_def()
        self._parallel = f.get_parallel()
        return node

    def apply_all(self, ast):
        """ Apply mutation to all matches 
            
            @return last node changed
        """
        start_node = None
        self.ast = ast
        f = OmpParallelForFilter(device = self.device)
        num = 0
        old_parallel = self._parallel
        try:
            for elem in f.iterate(ast):
                # Save previous state
                old_file_name = self.file_name
                self.file_name = self.file_name + str(num)
                old_clauses = self._clauses
                self._clauses = {}
                # Current scope variables
                self._func_def = f.get_func_def()
                self._parallel = f.get_parallel()
                # Apply mutation
                start_node = self.mutatorFunction(ast, elem)
                # Restore previous state
                self.file_name = old_file_name
                self._clauses = old_clauses
                num += 1
        except NodeNotFound as nf:
            print str(nf)
        except StopIteration:
            return self._parallel
        return start_node

    def mutatorFunction(self, ast, ompParallelFor_node):  
        """ Capture the clauses
        """
        # Solve the error when the pragma parallel has not clauses
        if ompParallelFor_node.clauses == None:
            ompParallelFor_node.clauses = {} 
        clause_dict = self._get_dict_from_clauses(ompParallelFor_node.clauses, self.ast)
        shared_params = clause_dict['SHARED']
        modified_shared_vars = clause_dict['COPY_OUT']
        copyin_shared_vars = clause_dict['COPY_IN']
        # If no copy clauses specified, default to shared params
        if len(copyin_shared_vars) == 0 and len(modified_shared_vars) == 0:
            modified_shared_vars = shared_params
            copyin_shared_vars = shared_params
        nowait = clause_dict.has_key('NOWAIT')
        private_params = clause_dict['PRIVATE']
        lastprivate_params = clause_dict['LASTPRIVATE']
        reduction_params = clause_dict['REDUCTION']
        # Capture the llc clauses
        llc_clause_dict = self._get_dict_from_llc_clauses(ompParallelFor_node.stmt, self.ast)
        result_params = llc_clause_dict['RESULT']
        nc_result_params = llc_clause_dict['NC_RESULT']
        ####### begin of mutator MPI_OmpFor ########
        """ MPI mutator, writes memory transfer operations for a parallel region
        """
        # Capture the values loop for init and cond.
        loop = ompParallelFor_node.stmt
        if result_params or nc_result_params:
            loop = loop.stmt
        # Inicialización del bucle
        loop_init = loop.init
        # Condición del bucle
        loop_cond = loop.cond
        # Actualización del bucle
        loop_next = loop.next
        # Valor inicial de la variable del bucle
        loop_initial = loop.init.rvalue
        # Valor final de la variable del bucle
        loop_finalice = str(loop.cond.right)
        if (loop.cond.op == '<'):
            loop_finalice += " - 1"
        # Variable del bucle
        loop_var = loop.init.lvalue
        # Bloque del bucle
        loop_stmt = loop.stmt
        subs_dir = dict()
        """ build the template_code """
        if ((loop_var) and (loop_initial) and (loop_finalice) and ((type(loop.next) == c_ast.UnaryOp) or (str(loop.next.rvalue) == '1'))):
            template_code = self.processLlcCodeUnit(variables = ompParallelFor_node.stmt, omp_cl = ompParallelFor_node.clauses, ast = self.ast, translate = subs_dir) 
        else:
            pass
        subs_dir.update({ 'loop_init': loop_init,
                           'loop_cond': loop_cond,
                           'loop_next': loop_next,
                           'loop_initial': loop_initial,
                           'loop_finalice': loop_finalice,
                           'loop_var': loop_var,
                           'loop_stmt': loop_stmt,
                           })
        forall_unit = self.fillTemplate(template_code, subs_dir, name = 'template forall unit', show = False)
        self.main_ast = getFileNode(ompParallelFor_node)
        source = mpi_outline_loop_stmts(ompParallelFor_node.stmt.stmt, file_name = self.file_name, template_source = forall_unit, ast = self.ast)
        genFile = C_FileType()
        self.source_storage.addFile(self.file_name, genFile)
        self.source_storage.append(self.file_name, genFile, '#include "llcomp_llc.h" \n')
        self.source_storage.append(self.file_name, genFile, source)
        updateSymbolTableFromNode(ast)

