# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

"""
Frontend Factory


@author: Ruyman Reyes Castro <rreyes@ull.es>

"""


import sys
import os
import config

import copy

from Frontend.Acc.plyparser import ParseError

class GeneralError(Exception):
    def __init__(self, str):
        self.str = str
        
    def __str__(self):
        print ":: " + str(self.str)


# These module-scope dictionaries holds the cache for the different parser/lexer/prepro classes
parser_cache = {}
lexer_cache = {}
prepro_cache = {}



class LanguageNotDefined(Exception):
    pass


class FrontendFactory(object):
    """ Dynamic creation of lexer and parser. 
    """

    def getInstalledFrontends(self):
        installed = set()
        for elem in os.listdir(config.FRONTEND_DIR):
            if os.path.isdir(os.path.join(config.FRONTEND_DIR, elem)):
                if os.path.isfile(os.path.join(config.FRONTEND_DIR,str(elem) + '/_' + str(elem).lower() + '_ast.cfg')):
                    installed.add(elem)
        return installed
    def checkLanguageList(self, languages):
        installed = self.getInstalledFrontends()
        for elem in languages:
            if elem not in installed:
                raise GeneralError('Language ' + str(elem) + " not installed. Try one of : " + str(installed))
        return True
       
    def getLexerFromLanguageList(self, languages):
        if "-".join(languages) in lexer_cache:
            return lexer_cache["-".join(languages)]
        # Check that we have modules for listed languages
        self.checkLanguageList(languages)
        lexer_class_tuple = ()
        tmp = copy.copy(languages)
        tmp.reverse()
        tokens = []
        states = tuple()
        for name in tmp:
#            print " Adding " + str(name) + " to language classes" 
            myclass = __import__(config.FRONTEND_MODULE + '.' + str(name) + "." + str(name).lower() + '_lexer', globals(), locals(), [str(name),])
            lexer_class = getattr(myclass, name + 'Lexer')
#            print "Class Tokens " + str(lexer_class.get_tokens()) + " Type : " + str(type(lexer_class.get_tokens()))
            tokens = lexer_class.get_tokens() + tokens
            states = lexer_class.get_states() + states
#            print "Total Tokens " + str(tokens)
            lexer_class_tuple += (lexer_class,)
        
        def get_tokens(self,):
            return tokens
        
        def get_states(self,):
            return states

#        print " Tokens for : " + str(languages)
#        print " are :"  + str(get_tokens(1))
#        import pdb
#        pdb.set_trace()
        my_lexer = type("CurrentLexer", lexer_class_tuple, {})
        static_get_tokens = staticmethod(get_tokens).__get__(None, my_lexer)
        static_get_states = staticmethod(get_states).__get__(None, my_lexer)
        setattr(my_lexer, 'get_tokens', static_get_tokens)
        setattr(my_lexer, 'get_states', static_get_states)

        lexer_cache["-".join(languages)] = my_lexer
        return my_lexer
       
       
    def getParserFromLanguageList(self, languages,):
        if "-".join(languages) in parser_cache:
            return parser_cache["-".join(languages)]
        # Check that we have modules for listed languages
        self.checkLanguageList(languages)
        parser_class_tuple = ()
        tmp = copy.copy(languages)
        tmp.reverse()
        for name in tmp:
            myclass = __import__(config.FRONTEND_MODULE + '.' + str(name) + "." + str(name).lower() + '_parser', globals(), locals(), [str(name),])
            parser_class = getattr(myclass, name + 'Parser')
            parser_class_tuple += (parser_class,)
        # print parser_class_tuple
        # Create the virtual class
        my_parser = type("CurrentParser", parser_class_tuple, {})
        parser_cache["-".join(languages)] = my_parser
        return my_parser
    
    def getPreproFromLanguageList(self, languages, includes=[]):
# Cannot use prepro cache with include list
        if len(languages) > 1:
            raise NotImplemented
        if "-".join(languages) in prepro_cache and (len(includes) != 0):
            return prepro_cache["-".join(languages)]
        self.checkLanguageList(languages)
        name = languages[0]
        myclass = __import__(config.FRONTEND_MODULE + "." + str(name) + "." + str(name).lower() + '_prepro', globals(), locals(), [str(name),])
        if len(includes) != 0:
            myclass.INCLUDEPATH = includes
        prepro_class = getattr(myclass, name + 'Prepro')
        prepro_cache["-".join(languages)] = prepro_class
        return prepro_class
    
    def parse_with_language_list(self, source, source_name, languages, includes=[]):
        if languages is None or len(languages) == 0:
            raise LanguageNotDefined
        # TODO: Implement a Flightweight pattern here , reuse the parsers when languages are the same!
        prepro_class = self.getPreproFromLanguageList(['C99',], includes)
        lexer_class = self.getLexerFromLanguageList(languages)
        parser_class = self.getParserFromLanguageList(languages,)
        prepro = prepro_class()
        if config.MULTIPLE_FILE_PREPRO:
            ## TODO: New implementation in preprocess
            stripped_source = prepro.preprocess2(source, includes)
        else:
            ## The following line is the old preprocessor
            stripped_source = prepro.preprocess(source)
        parser = parser_class(lex_optimize = config.LEX_OPTIMIZE, yacc_optimize = config.YACC_OPTIMIZE, lexer_class = lexer_class, yacc_debug=config.YACC_DEBUG)
        try:
            ast = parser.parse(stripped_source, source_name)
        except ParseError as pe:
            #import pdb
            #pdb.set_trace()
            code = stripped_source.split('\n')
            _window = range(0, len(code))
            filename = '<source_file>'
            try:
                errline = int(pe.args[0].split(':')[1]) 
                filename = pe.args[0].split(':')[0]
                w = 4; _window = range(max(errline-w-1, 0), min(errline+w, len(code)))
            except Exception as err:
                print err.message
            # Show parse error source line
            print filename, ":"
            for line in _window:
                print str(line+1) + ": " + code[line]
            # Show parse error message
            print "\nParseError:", pe.message, "\n"
            raise pe                
        return [ast, prepro_class, lexer_class, parser_class,]       

    def parse_reuse(self, source, source_name):
        # TODO: Implement a Flightweight pattern here , reuse the parsers when languages are the same!
        stripped_source = self._prepro.preprocess(source)
        try:
            ast = self._parser.parse(stripped_source, source_name)
        except ParseError as pe:
            #import pdb
            #pdb.set_trace()
            code = stripped_source.split('\n')
            _window = range(0, len(code))
            filename = '<source_file>'
            try:
                errline = int(pe.args[0].split(':')[1]) 
                filename = pe.args[0].split(':')[0]
                w = 4; _window = range(max(errline-w-1, 0), min(errline+w, len(code)))
            except Exception as err:
                print err.message
            # Show parse error source line
            print filename, ":"
            for line in _window:
                print str(line+1) + ": " + code[line]
            # Show parse error message
            print "\nParseError:", pe.message, "\n"
            raise pe                
        return ast


#from Frontend.InternalRepr import CompoundAnnotation

# TODO: IMPLEMENT THIS PRORPERLY
# TODO: Move this to Shortcuts
def partial_parse(source, source_name = "Partial", languages = ['C99',]):
        ff = FrontendFactory()
        c99_prepro_class = ff.getPreproFromLanguageList(languages)
        c99_lexer_class = ff.getLexerFromLanguageList(languages)
        c99_parser_class = ff.getParserFromLanguageList(languages)
        parser = c99_parser_class(lex_optimize = config.LEX_OPTIMIZE, yacc_optimize = config.YACC_OPTIMIZE, lexer_class = c99_lexer_class, yacc_debug=config.YACC_DEBUG)
        stripped_source = c99_prepro_class().preprocess("int main() { " + source + "; }")
        # print "Striped source:" + str(stripped_source)
        ast = parser.parse(stripped_source, source_name)
        # Snippets should be annotated with stmts and decls
        # CompoundAnnotation().visit(ast)
        
        return ast.ext[-1].body.stmts[-1]


