# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

'''
Created on Nov 29, 2010

:author: rreyes
'''

from Backends.Common.Visitors.GenericVisitors import UndefIDFilter,\
    FuncCallFilter, FuncDeclOfNameFilter

from Tools.Tree import NodeNotFound, getFileNode
from Frontend.Shortcuts import getCurrentLanguageAst
import config
from Frontend.C99 import c99_ast
c_ast = getCurrentLanguageAst()

from Frontend.SymbolTable import updateSymbolTableFromNode

from Backends.Common.Mutators.AbstractMutator import AbstractMutator
from Backends.Common.Visitors.GenericVisitors import GenericFilterVisitor
from Tools.Tree import NodeNotFound, NodeNotValid
from Frontend.SymbolTable import getSymbolTableFromNode
from Backends.Common.TemplateEngine.TemplateParser import TemplateParser, get_template_array
from Frontend.InternalRepr import AstToIR
from Tools.Tree import ReplaceTool, getContainerAttribute, getFileNode
import copy
from Backends.Common.Visitors.GenericVisitors import FuncDefOfNameFilter,\
    FuncCallFilter
from Tools.Tree import getFileNode, NodeNotFound
from Backends.Common.Mutators.AstSupport import PointerMutator
from Backends.Common.Mutators.AstSupport import ReferenceMutator

from Backends.C99.Writers.AccWriter import AccWriter
from Backends.C99.Writers.C99Writer import CWriter as C99Writer

class InstructionFlowBreakFilter(GenericFilterVisitor):
    """ Return the first matching flow break statement
    """   
    def __init__(self, prev_brother = None):
        FLOW_BREAK_STMTS = [c_ast.Return, c_ast.Break, c_ast.Goto]
        self._isNested = False
        
        def condition(node):
            if type(node) in [c_ast.Return, c_ast.Goto]:
                return True
            elif type(node) in [c_ast.Break] and not self._isNested:
                return True
            return False
                
        super(InstructionFlowBreakFilter, self).__init__(condition_func = condition, prev_brother = prev_brother)
        
    
    def visit_For(self, node, offset, ignore = []):
        self._isNested = True
        self.generic_visit(node.init, offset,)
        self.generic_visit(node.cond, offset,)
        self.generic_visit(node.next, offset,)
        self.generic_visit(node.stmt, offset,)
        self.isNested = False
    

class DeclarationOrUseOfPointersFilter(GenericFilterVisitor):
    """ Return the first declaration or use of a pointer
    """   
    def __init__(self, prev_brother = None):
        
        def condition(node):
            if type(node) == c_ast.PtrDecl:
                return True
            elif type(node) == c_ast.UnaryOp and node.op == '*':
                return True
            return False
                
        super(DeclarationOrUseOfPointersFilter, self).__init__(condition_func = condition, prev_brother = prev_brother)
      
def compound_has_got_pointers(block_node):
    """ Return true if exists the declaration or use of a pointer in the block node
    
        :param block_node: The group of nodes where the function will apply the filter
    """
    try:
        DeclarationOrUseOfPointersFilter().apply(block_node)
        return True
    except NodeNotFound:
        return False

def block_is_SESE(block_node):
    try:
        InstructionFlowBreakFilter().apply(block_node)
    except NodeNotFound:
        return True
    raise NodeNotValid("Not a SESE Block")


class ParameterStyle:
        C_POINTER = '*'
        CXX_REFERENCE = '&'

class OutlineMutator(AbstractMutator):
    """ Outline a code block into a separate function.
    
    Given the subtree reference to a SESE block (called *sese_block* from now on), the Outliner involves:

    1. Creating a function *outlined_function* whose body is a copy of *sese_block*
    2. Replace the appropriate occurrences of C in the original AST by a function call to *outlined_function*.

    Outlining might be seen as a trivial task, but some consideration must be taken into account.
    Special attention must be given to the type definitions of the variables used inside the block *sese_block*. 
    
    The outlining algorithm is language independent (as far as the language has support for subroutines), but 
    there are several syntactic details involved in the generated code that it is not possible to decouple the
    algorithm from the implementation and therefore, a Strategy pattern cannot be used.
    
    To leverage the language-dependency, two templates have been defined for the translation: _outline_function_template and
    _outline_call_template. The first is in charge of constructing the code related to the destination function, and the latter
    is devoted to the function call creation.
    
    This class is heavily used within the |yacf| framework and therefore some hooks have been added to increase the
    flexibility. This hooks are: 
    * Defining the outlined function name, by setting the name parameter at the constructor 
    * Applying the outline or not: Sometimes, the developer don't want to replace the original code, 
        but she wants to obtain the outlined function. Setting replace parameter to False will change
        the default behavior of the mutator by doing all except of the replacing. 
    
    """
    

    
    
    def __init__(self, name = "foo", replace = True, 
                 parameter_style = ParameterStyle.C_POINTER, 
                 outline_function_template = None, 
                 outline_call_template = None, change_code_block = True,
                 writer = C99Writer):
        """
            :param name: Name of the outlined function.
            :param boolean replace: Specify if *sese_block* will be replaced by the function call or not.
            :param ParameterStyle parameter_style: Parameter passing style to be used in the outlined function
        """
        self._name = name
        self.outlined_subtree = None
        self._replace = replace
        self._extra_decls = []
        self.prototype_subtree = None
        self._parameter_style = parameter_style;
        self._parameters = []
        self._change_code_block = change_code_block
        self._undeclFunctions = []
        self._writer = writer
        
        if not outline_function_template:
            if self._parameter_style == ParameterStyle.C_POINTER:
                self._outline_function_template = """
                 <%!
                       from Backends.Common.TemplateEngine.Functions import is_complex_type, is_array, is_struct, is_pointer
                 %>

		            %for var in typedef:
		               ${var};
		            %endfor
		
		            %for var in set(variables):
                        %if var.structDecl != "":
		                    ${var.structDecl}
                        %endif
		            %endfor
		
                    <%
                       my_parameter_array = []
                    %>
                           
                    %for var in parameters:
                      %if is_array(var.ptr):
                         <%
                           #if is_array(var.ptr.type):
                               my_parameter_array.append(str(var.declaration) + "*" + str(var))
                           #else:
                           #    my_parameter_array.append(str(var.declaration) + str(var))
                         %>
                      %elif is_pointer(var.ptr):
                         <%
                           my_parameter_array.append(str(var.declaration) + str(var))
                         %>
                      %else:
                         <%
                           my_parameter_array.append(str(var.declaration) + "*_" + str(var))
                         %>
                      %endif
                    %endfor

		              
		            void ${funcName} (
                        ${', '.join(str(var) for var in my_parameter_array)}    
		                );
		                
		            void ${funcName} (
                        ${', '.join(str(var) for var in my_parameter_array)}    
		            ) {
		                %for var in parameters:
                            %if not is_complex_type(var.ptr):
		                        ${str(var.declaration) + str(var) + " = " +  "*_" + str(var)};
                            %endif
		                %endfor   
		                ${code};
		                %for var in parameters:
                            %if not is_complex_type(var.ptr):
    		                    ${"*_" + str(var) + " = " + str(var)};
                            %endif
		                %endfor   
		            }
                """
            else:
                self._outline_function_template = """
		            %for var in variables:
		                ${var.typedef or ""}
		            %endfor
		
		            %for var in variables:
		                ${var.structDecl or ""}
		            %endfor
		
		              
		            void ${funcName} (
		                ${', '.join(str(var.declaration) + str(var) for var in parameters)}    
		                );
		                
		            void ${funcName} (
		            ${', '.join(str(var.declaration) + str(var) for var in parameters)}    
		            ) {
		                ${code};
		            }
                """

        else:
            self._outline_function_template = outline_function_template
        if not outline_call_template:
            if self._parameter_style == ParameterStyle.C_POINTER:
                self._outline_call_template = """
                 <%!
                       from Backends.Common.TemplateEngine.Functions import is_complex_type
                 %>
		            %for var in typedef:
		               ${var};
		            %endfor
            
                    %for var in parameters:
                        ${var.structDecl or ""}
                    %endfor

            
                    int fake () {
                    void ${funcName} (
                        <%
                            my_parameter_array = []
                        %>
                           
                        %for var in parameters:
                          %if not is_complex_type(var.ptr):
                            <%
                                my_parameter_array.append(str(var.type) + ' &' + str(var))
                            %>
                          %else:
                            <%
                                from Frontend.InternalRepr import AstToIR
                                from Backends.C99.Writers.C99Writer import CWriter as C99Writer
                                node = AstToIR(Writer = C99Writer, ast = var.ptr).deepcopySubtree(var.ptr,)
                                node.init = None
                                my_parameter_array.append(str(node))
                            %>
                          %endif
                        %endfor

                        ${', '.join(str(var) for var in my_parameter_array)}    
                    );
                    }
                """
            else:
                self._outline_call_template = """
             %for var in parameters:
                ${var.typedef or ""}
            %endfor
            
            %for var in parameters:
                ${var.structDecl or ""}
            %endfor

            
            int fake () {
            ${funcName} (
            ${', '.join('&' + str(var) for var in parameters)}    
            );
            }
                """
        else:
            self._outline_call_template = outline_call_template
        self._typedefList = []
        self._funcCallList = []
        super(OutlineMutator, self).__init__()
    
    @property
    def outline_call_template(self):
        return self._outline_call_template
    
    @property
    def outline_function_template(self):
        return self._outline_function_template
    
    def filter(self, ast):
        return ast
    
    def getOutlinedFunction(self):
        return self.outlined_subtree  
    
    def setOutlinedFunction(self, subtree):
        self.outlined_subtree = subtree
    
    def getOutlinedPrototype(self):
        return self.prototype_subtree

    def getOutlinedFunctionCall(self):
        return self._funccall

    def getTypedefList(self):
        return self._typedefList
    
    def getFuncCallList(self):
        """ Return the list of function calls within a block.
            This method is mantained only for compatibility with
            old-fashion backends (like MPI and/or pure Cuda). It
            returns a list of function declarations of the functions
            called inside the outlined function.

            .. deprecated::
                Use funcCalls property instead. 
        """
        called_list = []
        undeclared_list = []
        for elem in self._funcCallList:
            try:
                funcDecl = FuncDeclOfNameFilter(name = elem.name).apply(getFileNode(self.main_ast))
            except NodeNotFound:
                # Declaration could not be found
                undeclared_list.append(elem)
                continue
            called_list.append(funcDecl.parent) 
        return called_list #, undeclared_list]
 
    @property
    def funcCalls(self):
        """ List of function calls inside the kernel """
        return self._funcCallList 


    def getParameterList(self):
        return self._parameters  
    
    def getTargetBlock(self):
        return self._target
    

    def _declToParam (self, node):
        """ Create a param node from a variable declaration.

            :param node: Declaration to translate
            :return:  Parameter form of the declaration

        """
        param = node.__copy__()
        # Quick return if possible
        if isinstance(param.type, c_ast.PtrDecl):
            return param
        
        if self._parameter_style == ParameterStyle.C_POINTER:
            PointerMutator().apply(param)
        elif self._parameter_style == ParameterStyle.CXX_REFERENCE:
            ReferenceMutator().apply(param)

        return param
    
    def buildOutlineFunction(self,  org_decls, target):
        """ Build a tree with the outlined function """
        from Backends.Common.Visitors.GenericVisitors import IDFilter
        from Tools.Tree import getContainerAttribute
        from Tools.Tree import ReplaceTool

        decls = [ elem.node.__copy__() for elem in org_decls]
        # Transform the declaration list into a parameter declaration
        params = [ self._declToParam(elem) for elem in decls]
        # Change variable access to value of variable access
        st = getSymbolTableFromNode(decls[0])

        if self._change_code_block:
		raise NotImplemented

#                for elem in decls:
#                    nodes_to_change = []
#                    # TODO: Put in one loop
#                    for node in IDFilter(id = c_ast.ID(name = elem.name,)).dfs_iter(target):
#                        #===============================================================
#                        # import pdb
#                        # pdb.set_trace()
#                        #===============================================================
#                        symbol = st.lookUp(node)
#                        if not isinstance(symbol.type, c_ast.ArrayDecl, ) and not isinstance(symbol.type, c_ast.PtrDecl):
#                            nodes_to_change.append(node)
#                    
#                    for id in nodes_to_change:
#                        value_node = None
#                        if self._parameter_style == ParameterStyle.C_POINTER:
#                            value_node = c_ast.ExprList(exprs = [c_ast.UnaryOp(op = '*', expr = id,)])
#                            value_node.exprs[0].parent = value_node
#                        elif self._parameter_style == ParameterStyle.CXX_REFERENCE:
#                            value_node = id
#                        else:
#                            raise Exception
#
#                        ReplaceTool(new_node = value_node, old_node = id).apply(id.parent, getContainerAttribute(id))
#                        from Frontend.SymbolTable import updateSymbolTableFromNode
#                        updateSymbolTableFromNode(self.main_ast)
#
        
        #  print "****************************"

        self._parameters =  get_template_array(org_decls, self.main_ast, st = st)
        decls_tmp_array  =  get_template_array(org_decls, self.main_ast, st = st)
        decls_tdef_array =  get_template_array(self._decls_typedef, self.main_ast, st = st)
        subs_dir = {'variables' : decls_tmp_array , 
                        'parameters' : self._parameters, 
                        'code' : str(target), 
                        'funcName' : self._name,
                        'ArrayDecl' : str(c_ast.ArrayDecl),
                        'PtrDecl' : str(c_ast.PtrDecl),
                        'typedef': self.get_typeDefinitions(decls_tdef_array, self.main_ast),}
        template_code = TemplateParser(self.outline_function_template).render(**subs_dir)
        subtree = self.parse_snippet(template_code, subs_dir, 'fbody', show = False, languages = config.CURRENT_LANGUAGES)
        #print "****************************"
        #import pdb
        #pdb.set_trace()
        return [subtree.ext[-1], subtree.ext[-2]]

    def get_typeDefinitions(self, declarations, main_ast):
        """ Build a list with all typedefs used in declarations or parameters.
        """
        # declarations = get_template_array(decls_tmp_array, main_ast, st)
        typedef = list()
        for decl in declarations:
            if decl.typedef and not decl.typedef in typedef:
                typedef.append(decl.typedef)
        return typedef
        
    def get_typeDefinitions_node(self, declarations, main_ast):
        """ Build a list with all typedefs used in declarations or parameters.
        """
        # declarations = get_template_array(decls_tmp_array, main_ast, st)
        typedef = list()
        typedef_n = list()
        for decl in declarations:
            if decl.typedef and not decl.typedef in typedef:
                typedef_n.append(decl.typedef_node)
                typedef.append(decl.typedef)
        return typedef_n

    def add_additional_decls(self, decl_list):
        """ The declarations in the decl_list are added as extra parameters to the 
                outlined function. 
        """
        self._extra_decls += decl_list


    def buildFunctionTypedefList(self, params, target):
        """ Build a list with the typedefs required for the outlined function. """
        # Outline template        
	params_tmp_array = get_template_array(params, self.main_ast)
        decls_tdef_array =  get_template_array(self._decls_typedef, self.main_ast)
        subs_dir = {'parameters' : params_tmp_array,
                    'code' : str(target), 
                    'funcName' : self._name, 
                    'typedef': self.get_typeDefinitions(decls_tdef_array, self.main_ast),}
        template_code = TemplateParser(self.outline_call_template).render(**subs_dir)
        return self.parse_snippet(template_code, subs_dir, 'tdef', show = False).ext[0:-1]
    
    def buildOutlineCall(self, params, target):
        """ Build a function call to the outlined function.
        
            :param params: Params of the function
            :return: Subtree with the function call """
	return None
        # Outline template
	params_tmp_array = get_template_array(params, self.main_ast)
        subs_dir = {'parameters' : params_tmp_array, 
                    'funcName' : self._name, 
                    'ArrayDecl': str(c_ast.ArrayDecl),
                    'typedef': self.get_typeDefinitions(params_tmp_array, self.main_ast),}
        template_code = TemplateParser(self.outline_call_template).render(**subs_dir)
        return self.parse_snippet(template_code, subs_dir, 'fcall', show = False).ext[-1].body.block_items[-1]

    def _merge_decls_typedef(self, destination_list, decls):
        """ Returns first level Typedef decls given a list of decls and a node to explore"""
        # Typedef decl need to be in order from simple to complex types
        decls_typedef = destination_list
        if len(set(decls_typedef).union(set(decls))) > len(decls_typedef):
            decls_typedef = decls + decls_typedef
        else:
            if decls:
                # If typedef already exists move to head

                _up = False
                for decl in decls:
                    index = decls_typedef.index(decl)
                    if index > 0:
                        del(decls_typedef[index])
                        _up = True
                if _up:
                    decls_typedef = decls + decls_typedef
        return decls_typedef
    
    def _get_all_decls_typedef(self, decls, sese_block):
        """ """
        root = sese_block.getRootNode()
        
        # This method search every Decl. More than filter only IDs
        true_decls = self._getUndeclDeclSymbols(sese_block,)
        # Merge respecting typedef dependencies
        decls_typedef = self._merge_decls_typedef(true_decls, decls)
        
        params_tmp_array = []
        _len_decls = len(decls_typedef) + 1
        # Add typedef decls until there is no more decls dependencies required
        while len(decls_typedef) != _len_decls:
            _len_decls = len(decls_typedef)
            # Get list of typedef decls with type TemplateVarNode
            params_tmp_array = get_template_array(decls_typedef, root)
            # Get node list of typedef decls
            _typedefNodeList = self.get_typeDefinitions_node(params_tmp_array, root)
            # Look for every undecl symbol in _typedefNodeList
            for typedefNode in reversed(_typedefNodeList):
                # Look for type definitions given typedefNode
                tmp = self._getUndeclTypeSymbols(typedefNode,)
                # Merge again decls_typedef with new decls_typedef(tmp)
                decls_typedef = self._merge_decls_typedef(decls_typedef, tmp)
        return [_typedefNodeList, decls_typedef]
    
    def mutatorFunction(self, sese_block):
        """
             Apply the OutlineMutation to the given sese_block.
             The following class attributed are modified after the mutation:
             * _funccall
             * _typedefList
             * _funcCallList
             * outlined_subtree
             * prototype_subtree


             :param sese_block: SESE block to outline from the code. Parent attribute must be valid.
             :return: Replaces sese_block with a function call to the newly created function. 
        """
        # Get the variables declared outside this scope 
        # outside_scope_vars = self._getUndeclVars(sese_block,)
        #import pdb
        #pdb.set_trace()
        outside_scope_symbols = self._getUndeclSymbols(sese_block,)
        outside_scope_nodes = self._getUndeclNodes(sese_block,)
        
        # Get the declarations of those variables
        decls = []
        if (len(outside_scope_symbols) > 0):
            decls = outside_scope_symbols # self._getDeclList(outside_scope_nodes)
        if len(self._extra_decls):
            decls = list(set(decls).union(set(self._extra_decls)))

        # Get all typedef declaration needed for sese_block
        [self._typedefNodeList, self._decls_typedef] = self._get_all_decls_typedef(decls, sese_block)
        
        # Clone the subtree
        target = AstToIR(Writer = self._writer, ast = sese_block).deepcopySubtree(sese_block,)
        
        # Make new envelope function foo
        foo = c_ast.FuncDef(decl = c_ast.Decl(name = 'foo',
                                              quals = [], storage=[], funcspec=[],
                                              type=c_ast.TypeDecl(declname='foo', quals=[],
                                                  type=c_ast.IdentifierType(names=['void'])),
                                              init=None, bitsize=None),
                            param_decls = None,
                            body = target)
        target.parent = foo
        
        # Build new FileAst with outlined function
        nFAST = c_ast.FileAST(ext = self._typedefNodeList + [foo])
        foo.parent = nFAST
        # Update its SymbolTable
        updateSymbolTableFromNode(nFAST)
        #AstToIR.link_subtree_parents(root)
        
        self.main_ast = getFileNode(target)
        #AstToIR.link_subtree_parents(target)

        # Build the function
        [self.outlined_subtree, self.prototype_subtree] = self.buildOutlineFunction(decls, target)
        # Build the function call
        self._funccall = self.buildOutlineCall(decls, target)
        # Build list of required typedefs       
        self._typedefList = self.buildFunctionTypedefList(decls, target)
        # Build list of function calls inside Kernel
        self._funcCallList = self._getFuncCallList(target)
        # Build new target block SESE
        self._target = target        
        # Replace the old compound stmt with a function call
        if self._replace:
            ReplaceTool(new_node = self._funccall, old_node = sese_block).apply(sese_block.parent, getContainerAttribute(sese_block))

        #nFAST = c_ast.FileAST(ext = [self.getOutlinedFunction()])
        #target.parent = nFAST
 

def inline_called_functions(non_visited_funcalls, objetive, 
                    native_list, allow_native):
        """ Inline called functions 
        """
        undeclared_list = []
        visited = []
        called_list = []
        while len(non_visited_funcalls) > 0:
            try:
                elem = non_visited_funcalls.pop()
                if not allow_native or (allow_native and not elem.name in native_list):
                    #~ print " Extracting function named: " + str(elem.name)
                    funcDef = FuncDefOfNameFilter  (name = elem.name).apply(getFileNode(objetive))
                    called_list.append(funcDef) 
                    visited.append(elem.name)
                    # Check for nested calls
                    for elem in FuncCallFilter().iterate(funcDef.body):
                        if not elem in visited:
                            non_visited_funcalls.append(elem)
                elif allow_native and elem.name in native_list:
                    print " Native call to :" + str(elem.name)
                else:
                    print " Function " + elem.name + " is neither native or declared "
                    raise NotImplemented
            except NodeNotFound:
                # Declaration could not be found
                undeclared_list.append(elem)
                continue
        if len(undeclared_list) > 0:
            print " Following functions were called but not declared: "
            print [str(elem.name) for elem in undeclared_list]
	    print " Will assume they are native "
	# Reverse list to match declaration order
        called_list.reverse()
        return called_list
        

def c_outline_loop_stmts(objetive, om):
    """
        Outline a C statement from a For Loop
        
    """
    # Abort if not a SESE block   
    assert block_is_SESE(objetive)
    
    
    c_template = """
    %for typedef in typedeflist:
        ${typedef};
    %endfor
    
    %for func in funccalllist:
        ${func.decl};
    %endfor
    
       
    ${kernelProto};
    
    ${kernelFunction};
    
    
    """
    typedeflist = om.getTypedefList()

    
    allow_native = True
    native_list =  ['sqrt', 'min', 'max', 'log', 'fabs', 'pow',]
    non_visited_funcalls = om.funcCalls[:]

    called_list = inline_called_functions(non_visited_funcalls, objetive, 
		                               native_list, allow_native)
    outlinedProto = om.getOutlinedPrototype()    
    outlinedFunction = om.getOutlinedFunction()
    #print "outlinedFunction:", outlinedFunction

    # Remove attributes
    for f in called_list:
        setattr(f.decl, "storage",  ['extern',])

    #===========================================================================
    from mako.template import Template
    kernel_source = Template(c_template).render_unicode(typedeflist = om.getTypedefList(),
                                                             funccalllist = called_list,
                                                             outlinedProto = outlinedProto,
                                                             kernelProto = outlinedProto,
                                                             kernelFunction = outlinedFunction,
                                                            )    
    # Restore attributes
    for f in called_list:
        # static cannot be applied anymore to that method, as it would be linked later
        setattr(f.decl, "storage",  [])


 
    return kernel_source




