# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

from Frontend.C import c_parser, c_ast


from Backends.Common.Visitors.GenericVisitors import * 
from Backends.Common.Mutators.AbstractMutator import *

from Tools.Tree import InsertTool, RemoveTool, ReplaceTool, getContainerAttribute, \
			NodeNotFound
from Frontend.Shortcuts import getCurrentLanguageAst
c_ast = getCurrentLanguageAst()

from Frontend.Omp import omp_ast

from MiddleEnd.Loop.Analysis import ParametrizeLoopTool
from MiddleEnd.Loop.Common import LoopFilter, IDFilter, \
		    ParameterMustBeDefinedException, \
                   PragmaReverseFilter, \
                   CompundFilter

class LoopInterchangeFilter(GenericFilterVisitor):
    """ Returns the first node with the given attribute
    """
    def __init__(self):
        def condition(node):
            if type(node) == omp_ast.llcInterchange:
                return True
            return False
        super(LoopInterchangeFilter, self).__init__(condition_func = condition)

class LoopInterchangeMutator(AbstractMutator):
    """ Apply Loop Interchange """ 
    def __init__(self, debug=False):
        self.debug = debug
        super(LoopInterchangeMutator, self).__init__()

    def filter(self, ast):
        """  """
        raise NotImplemented

    def filter_iterator(self, ast):
        """ Fast filter  """
        return NotImplemented

    def fast_filter(self, ast):
        """ Fast filter , looking for binary expressions """
        return LoopInterchangeFilter().dfs_iter(ast)

    def setIndexes(self, loopId_1, loopId_2):
	""" Set self.index_1 and self.index_2 if parameters are given"""
	if loopId_1: self.index_1 = loopId_1
	else: raise ParameterMustBeDefinedException('loopId_1')
	if loopId_2: self.index_2 = loopId_2
	else: raise ParameterMustBeDefinedException('loopId_2')
        
    def mutatorFunction(self, ast, loopId_1 = None, loopId_2 = None):
        """ Mutator code 

            :param ast: node returned by the filter. It can be given directly
            :param loopId_1: optional. Name of 1st loop index to interchange
            :param loopId_1: optional. Name of 2nd loop index to interchange
        """

        from MiddleEnd.Loop.Mutators.LoopStripMining import get_old_index_name
        if (get_old_index_name(str(loopId_1))
		== get_old_index_name(str(loopId_2))):
	    return ast

	print "*** Applying INTERCHANGE over loops (" + \
		str(loopId_1) + ", " + \
		str(loopId_2) + ") ***"
        
        # Test if ast points to a c_ast.For node
        if isinstance(ast, omp_ast.llcInterchange):
	    interchange_node = ast
	    assert isinstance(interchange_node.loop, c_ast.For)
	else:
	    interchange_node = None
	    assert isinstance(ast, c_ast.For)

        # Get and Look for index strings of loops to interchange
	# if they are both defined 
        if interchange_node:
	    try:
		self.index_1 = interchange_node.loopId_1
		self.index_2 = interchange_node.loopId_2
	    except AttributeError:
		if (loopId_1 or loopId_2):
		    self.setIndexes(loopId_1, loopId_2)
		else:
		    self.index_1 = None
		    self.index_2 = None
	elif (loopId_1 or loopId_2):
	    self.setIndexes(loopId_1, loopId_2)
	else:
	    self.index_1 = None
	    self.index_2 = None

	# Search IDs for their index strings
	id_1 = IDFilter(self.index_1).apply(ast) if self.index_1 else None
	#print self.index_1, type(self.index_1)
	id_2 = IDFilter(self.index_2).apply(ast) if self.index_2 else None
        #print self.index_2, type(self.index_2)
	
	# If self.index_1 and self.index_2 exist it swap loops whose
	#     iterating variables are self.index_1 and self.index_2
	# Else it swap first and second loops
        first_loop = LoopFilter(id_1).apply(ast)
        assert isinstance(first_loop, c_ast.For)
        try:
	    second_loop = LoopFilter(id_2).apply(first_loop.stmt)
	except NodeNotFound:
	    second_loop = LoopFilter(id_2).apply(first_loop)
        assert isinstance(second_loop, c_ast.For)

        
        # Look for a pragma tag with each loop and atach it to interchange
        # loops with their respectives pragmas
        first_pragma_loop = None
        try:
            first_pragma = PragmaReverseFilter().apply(first_loop)
            first_pragma_loop = LoopFilter().apply(first_pragma)
        except NodeNotFound:
            pass            
        if (first_pragma_loop == first_loop):
            first_loop = first_pragma
            
        second_pragma_loop = None
        try:
            second_pragma = PragmaReverseFilter().apply(second_loop)
            second_pragma_loop = LoopFilter().apply(second_pragma)
        except NodeNotFound:
            pass            
        if (second_pragma_loop == second_loop):
            second_loop = second_pragma
            
        #self.print_("first_pragma_loop", first_pragma_loop)
        if self.debug: self.print_("first_loop", first_loop)
        #self.print_("second_pragma_loop", second_pragma_loop)
        if self.debug: self.print_("second_loop", second_loop)
        if self.debug: raw_input(" ********************** Press ENTER to continue ********************** ")
        
        
        #loop1_parameters = ParametrizeLoopTool().apply(first_loop,)
        #loop2_parameters = ParametrizeLoopTool().apply(second_loop,)

        #------------------------------------------------------------
        #from Tools.Debug import DotDebugTool
        #DotDebugTool(highlight = [first_loop, ]).apply(ast)
        #------------------------------------------------------------

        # ==============================================================
        # A. Extract second loop from first loop childs
        # ==============================================================
        second_loop_old_parent = getattr(second_loop, 'parent')
        att_second_loop = getContainerAttribute(second_loop) # 'block_items'
        second_loop_parent_block = getattr(second_loop.parent, \
                                                att_second_loop)

        if (type(second_loop_parent_block) == type([])):
            index_loop2 = second_loop_parent_block.index(second_loop)
            second_loop_parent_block[index_loop2] = \
                                                    c_ast.ID('PlaceHolder')
        else:
            setattr(second_loop.parent, att_second_loop, c_ast.ID('PlaceHolder'))
            print "WARNING: second_loop_parent_block is not a list"

        setattr(second_loop, 'parent', first_loop.parent)
            
        if self.debug: print "************************ A ***************************"        
        if self.debug: self.print_("att_second_loop", att_second_loop) # block_items
        if self.debug: self.print_("index_loop2", index_loop2)
        if self.debug: self.print_("second_loop_parent_block[index_loop2]", second_loop_parent_block[index_loop2])
        
        if self.debug: self.print_("second_loop", second_loop)
        #if self.debug: self.print_("second_loop.parent", second_loop.parent)
        
        if self.debug: self.print_("first_loop", first_loop)
        if self.debug: self.print_("first_loop.parent", first_loop.parent)
        
        if self.debug: self.print_("second_loop_old_parent", second_loop_old_parent)
        if self.debug: raw_input(" ********************** Press ENTER to continue ********************** ")

        # ==============================================================
        # B. Move second loop up to the root of first loop
        # ==============================================================
        first_loop_old_parent = first_loop.parent
        att_first_loop = getContainerAttribute(first_loop)
        first_loop_parent_block = getattr(first_loop.parent, att_first_loop)        
	
        if (type(first_loop_parent_block) == type([])):
            index_loop1 = first_loop_parent_block.index(first_loop)
            first_loop_parent_block[index_loop1] = second_loop
        else:
            print "WARNING: first_loop_parent_block is not a list"
            setattr(first_loop.parent, att_first_loop, second_loop)
	
        #setattr(first_loop, 'parent', second_loop_old_parent)
        
	
        if self.debug: print "************************ B ***************************"        
        if self.debug: self.print_("att_first_loop", att_first_loop)
        if self.debug: self.print_("index_loop1", index_loop1)
        if self.debug: self.print_("first_loop_parent_block[index_loop1]", first_loop_parent_block[index_loop1])
        
        if self.debug: self.print_("first_loop", first_loop)
        if self.debug: self.print_("first_loop_old_parent", first_loop_old_parent)
        if self.debug: raw_input(" ********************** Press ENTER to continue ********************** ")

        #------------------------------------------------------------
        #from Tools.Debug import DotDebugTool
        #DotDebugTool(highlight = [first_loop, second_loop]).apply(first_loop_old_parent.parent)
        #------------------------------------------------------------      
        
        # ==============================================================
        # C. Swap second loop statements with first loop statements
        # ==============================================================
        second_loop_block = CompundFilter().apply(second_loop)
        if (second_loop.stmt == second_loop_block):
            second_loop_block_parent = second_loop
        else:
            second_loop_block_parent = second_loop_block.parent
        
        first__loop_block = CompundFilter().apply(first_loop)
        if (first_loop.stmt == first__loop_block):
            first__loop_block_parent = first_loop
        else:
            first__loop_block_parent = first__loop_block.parent
        
        if self.debug: print "************************ pre-C ***************************"        
        if self.debug: self.print_("first__loop_block.parent", first__loop_block.parent)
        if self.debug: self.print_("first__loop_block", first__loop_block)        
        if self.debug: self.print_("first_loop", first_loop)
        
        if self.debug: self.print_("second_loop", second_loop)
        if self.debug: self.print_("second_loop_block", second_loop_block)
        if self.debug: self.print_("second_loop_block.parent", second_loop_block.parent)
        if self.debug: raw_input(" ********************** Press ENTER to continue ********************** ")
        
        setattr(second_loop_block_parent, 'stmt', first__loop_block)
        setattr(first__loop_block_parent, 'stmt', second_loop_block)
        setattr(first__loop_block, 'parent', second_loop)

        if self.debug: print "************************ C ***************************"
        if self.debug: self.print_("first_loop", first_loop)
        if self.debug: self.print_("first_loop_old_parent", first_loop_old_parent)
        if self.debug: self.print_("first__loop_block", first__loop_block)
        if self.debug: self.print_("first__loop_block_parent", first__loop_block_parent)
        
        if self.debug: self.print_("second_loop", second_loop)
        if self.debug: self.print_("second_loop_block", second_loop_block)
        if self.debug: self.print_("second_loop_block_parent", second_loop_block_parent)
        if self.debug: raw_input(" ********************** Press ENTER to continue ********************** ")

        #------------------------------------------------------------
        #from Tools.Debug import DotDebugTool
        #DotDebugTool(highlight = [first_loop, second_loop]).apply(first__loop_block)
        #DotDebugTool(highlight = [first_loop, second_loop]).apply(second_loop_block)
        #------------------------------------------------------------      
        
        # ==============================================================
        # D. Reconstruct new first loop ( old second_loop )
        # ==============================================================
   
        if self.debug: print "************************ pre - D ***************************"        
        if self.debug: self.print_("second_loop_old_parent", second_loop_old_parent)
        if self.debug: self.print_("first_loop", first_loop)
        if self.debug: self.print_("first_loop.parent", first_loop.parent)
        if self.debug: self.print_("index_loop1", index_loop1)
        if self.debug: self.print_("first_loop_parent_block[index_loop1]", first_loop_parent_block[index_loop1])
        
        if self.debug: self.print_("second_loop", second_loop)
        if self.debug:  self.print_("index_loop2", index_loop2)
        if self.debug: self.print_("second_loop_parent_block[index_loop2]", second_loop_parent_block[index_loop2])
        
        #self.print_("ast", ast)
        if self.debug: raw_input(" ********************** Press ENTER to continue ********************** ")
   
        if (type(second_loop_parent_block) == type([])):
            second_loop_parent_block[index_loop2] = first_loop            
        else:
            setattr(second_loop_old_parent, att_second_loop, first_loop)
            print "WARNING: second_loop_parent_block is not a list"
        
        if (type(first_loop_parent_block) == type([])):
            first_loop_parent_block[index_loop1] = second_loop
        else:
            print "WARNING: first_loop_parent_block is not a list"
            setattr(first_loop.parent, att_first_loop, second_loop)
        
        setattr(first_loop, 'parent', second_loop_old_parent)
        second_loop_loop = LoopFilter().apply(second_loop_old_parent.parent)        
        setattr(first_loop.parent, 'parent', second_loop_loop)

        ast = second_loop
        
        #------------------------------------------------------------
        #from Tools.Debug import DotDebugTool
        #DotDebugTool(highlight = [first_loop, ]).apply(first_loop.parent)
        #------------------------------------------------------------  
	
        if self.debug: print "************************ D ***************************"
        if self.debug: self.print_("first_loop", first_loop)
        if self.debug: self.print_("second_loop", second_loop)
        #self.print_("second_loop_parent_block", second_loop_parent_block)
        if self.debug: self.print_("second_loop_old_parent", second_loop_old_parent)
        if self.debug: self.print_("first_loop.parent", first_loop.parent)
        if self.debug: self.print_("first_loop.parent.parent", first_loop.parent.parent)
        #self.print_("first_loop_parent_block", first_loop_parent_block)
        #self.print_("ast", ast)
        if self.debug: raw_input(" ********************** Press ENTER to continue ********************** ")

        #import sys
        #sys.exit()
	
        # Finally, replace old first loop with old second loop	
        #ReplaceTool(new_node = second_loop, old_node = first_loop).apply(first_loop.parent, getContainerAttribute(first_loop))	
        # TODO: Replace should update parent links?
        # TODO: Replace should support unknown parent attributes

        # ==============================================================
        # E. Remove pragma if exist ( pragma compatibility )
        # ==============================================================
        #if isinstance(interchange_node, omp_ast.llcInterchange):
	
        ## Note that the parent of any llc node is a pragma node
            #interchange_parent = ast.parent.parent		# pragma node
	
        #interchange_pragma = ast.parent			# llc node
        
            #node = interchange_parent.block_items
            #node[node.index(interchange_pragma)] = interchange_node.loop
            ##DotDebugTool(highlight = [first_loop, second_loop]).apply(interchange_parent)

            #del interchange_node
        
        # Return the new outer loop
        return ast

        
    def print_(self, name, node):
	print "\n", name, "(" + str(type(node)) + "):", node, "\n"

