# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------


from Frontend.Shortcuts import getCurrentLanguageAst

c_ast = getCurrentLanguageAst()

from Frontend import SymbolTable


from Tools.Tree import NodeNotFound

from Backends.Common.Visitors.GenericVisitors import DeclFilter

def reverse_decl_of_id(id, uplimit):
    """Returns the Decl node of a given ID node 
    
    :param id: ID Node
    :param uplimit: Parent pointer to stop reverse search
    
    :return: Decl node of the identifier
    """
    act = id.parent
    decl = None
    
    while True:
        # Check for a decl of the var as parameter
        if isinstance(act, c_ast.FuncDef):
            if hasattr(act.decl.type, 'args') and act.decl.type.args and act.decl.type.args.params:
                for elem in act.decl.type.args.params:
                    try:
                        decl = DeclFilter(attribute = "name", value = id.name).apply(elem)
                    except NodeNotFound:
                        pass
        # Taking advantage of lazy boolean evaluation, if the first part is false, 
        #     the second is not executed, so, we won't have an exception
        if hasattr(act, 'decls') and act.decls:
            for elem in act.decls:
                try:
                    decl = DeclFilter(attribute = "name", value = id.name).apply(elem)
                except NodeNotFound:
                    pass
                  
        # If we visit a outer Compound it filter looking for Decl
        if type(act) == c_ast.Compound:
            try:
                decl = DeclFilter(attribute = "name", value = id.name).apply(act)
            except NodeNotFound:
                pass
                  
        # The uppermost level is the FileAST, where decls are on the ext attribute
        if hasattr(act, 'ext') and act.ext:
            for elem in act.ext:
                try:
                    decl = DeclFilter(attribute = "name", value = id.name).apply(elem)
                except NodeNotFound:
                    pass
    
        if type(decl) == c_ast.Decl:
            return decl
        
        if act == uplimit or act == c_ast.FileAST or not hasattr(act,'parent'):
            break
    
        # import pdb
        # pdb.set_trace()
        # Keep going up
        act = act.parent
    
    # TODO: Raise identifier not declared
    return None




def decl_of_id(identifier, ast):
    """ Returns the declaration of an identifier.

        If the identifier is not in the symbol table, the table is rebuilt. If after being rebuilt, 
        it still missing, a reverse search is started.

        :param identifier: Identifier whose declaration want to be retrieved
        :param ast: Pointer to the FileAST of the code
        :rtype: Declaration node
    """
    from Frontend.InternalRepr import AstToIR
    from Backends.C99.Writers.OmpWriter import OmpWriter
    migrator = AstToIR(Writer = OmpWriter, ast = ast)
    # Transform not needed because it has been cached 
    st = migrator.symbolTable
    try:
         return st.lookUp(identifier).node
    except SymbolTable.IdentifierNotFound:
             try:
                 migrator.annotate()
                 return st.lookUp(identifier).node
             except SymbolTable.IdentifierNotFound:
                 return reverse_decl_of_id(identifier, ast)


def type_of_id(id, ast):
    """Returns the TypeDecl node of a given ID node 
    
            Calls decl_of_id but returns the type argument
    
            :rtype: TypeDecl node of a given ID node 
    
    """
    return decl_of_id(id,ast).type


#
#from Backends.Common.Writers.CommonWriter import OffsetNodeVisitor
#
#class NodeTypeVisitor(OffsetNodeVisitor):
#    # Priority among C types
#    _type_priority = ['long double', 'double','float', 'long int', 'int', 'short int']
#
#
#    def __init__(self, ast = None):
#        self._ast = ast
#
#    def get_type_name(self, elem):
#        """ Return a list of names for a type """
#        while not hasattr(type, 'type') and not hasattr(type, 'name'):
#            type = type.type
#
#        type = type_of_id(elem, self._ast)
#        while not hasattr(type, 'names') and not hasattr(type, 'name'):
#            type = type.type
#        if hasattr(type, 'names'):
#            return ' '.join(type.names)
#        else:
#            return str(type.name)
#
#
#    def coherce_type(self, typeA, typeB):
#        if self._type_priority.index(self.get_type_name(typeA)) > self._type_priority.index(self.get_type_name(typeB)):
#            return typeA
#        else:
#            return typeB
#
#    def visit_BinaryOp(self, node, offset = 0):
#        typeA = self.visit(node.left)
#        typeB = self.visit(node.right)
#        return self.coherce_type(typeA, typeB)
#
#
#
#def type_of_node(node, ast):
#          """Returns the TypeDecl node of an expression
#
#
#                Rules are: double>float>int
#
#                :rtype: TypeDecl node of a given expression node
#
#          """
#          return NodeTypeVisitor(ast = ast).visit(node)
#
